class PageMissingError(Exception):
    def __init__(self, title, xml):
        self.title = title
        self.xml = xml

    def __str__(self):
        return "page '%s' not found" % self.title


class ExportAbortedError(Exception):
    def __init__(self, index):
        self.index = index

    def __str__(self):
        return "Export from '%s' did not return anything." % self.index


class FileSizeError(Exception):
    def __init__(self, file: str, got_size: int, excpected_size: int, online_url: Optional[str] = None):
        self.file = file
        self.got_size = got_size
        self.excpected_size = excpected_size
        self.online_url = online_url

    def __str__(self):
        return f"File '{self.file}' size {self.got_size} is not match '{self.excpected_size}'." \
            + (f"(url: {self.online_url})" if self.online_url else "")


class FileSha1Error(Exception):
    def __init__(self, file, excpected_sha1):
        self.file = file
        self.excpected_sha1 = excpected_sha1

    def __str__(self):
        return "File '%s' sha1 is not match '%s'." % (self.file, self.excpected_sha1)



R_NEXT = r"(?<=&amp;dir=prev)&amp;offset=(?P<offset>\d+)"

REGEX_CANDIDATES = [
    # [0]
    # archiveteam 1.15.1 <td class="TablePager_col_img_name"><a href="/index.php?title=File:Yahoovideo.jpg" title="File:Yahoovideo.jpg">Yahoovideo.jpg</a> (<a href="/images/2/2b/Yahoovideo.jpg">file</a>)</td>
    # wikanda 1.15.5 <td class="TablePager_col_img_user_text"><a
    # href="/w/index.php?title=Usuario:Fernandocg&amp;action=edit&amp;redlink=1"
    # class="new" title="Usuario:Fernandocg (página no
    # existe)">Fernandocg</a></td>
    r'(?im)<td class="TablePager_col_img_name"><a href[^>]+title="[^:>]+:(?P<filename>[^>]+)">[^<]+</a>[^<]+<a href="(?P<url>[^>]+/[^>/]+)">[^<]+</a>[^<]+</td>\s*<td class="TablePager_col_img_user_text"><a[^>]+>(?P<uploader>[^<]+)</a></td>'

    # [1]
    # wikijuegos 1.9.5
    # http://softwarelibre.uca.es/wikijuegos/Especial:Imagelist old
    # mediawiki version
    ,r'(?im)<td class="TablePager_col_links"><a href[^>]+title="[^:>]+:(?P<filename>[^>]+)">[^<]+</a>[^<]+<a href="(?P<url>[^>]+/[^>/]+)">[^<]+</a></td>\s*<td class="TablePager_col_img_timestamp">[^<]+</td>\s*<td class="TablePager_col_img_name">[^<]+</td>\s*<td class="TablePager_col_img_user_text"><a[^>]+>(?P<uploader>[^<]+)</a></td>'

    # [2]
    # gentoowiki 1.18
    ,r'(?im)<td class="TablePager_col_img_name"><a[^>]+title="[^:>]+:(?P<filename>[^>]+)">[^<]+</a>[^<]+<a href="(?P<url>[^>]+)">[^<]+</a>[^<]+</td><td class="TablePager_col_thumb"><a[^>]+><img[^>]+></a></td><td class="TablePager_col_img_size">[^<]+</td><td class="TablePager_col_img_user_text"><a[^>]+>(?P<uploader>[^<]+)</a></td>'

    # [3]
    # http://www.memoryarchive.org/en/index.php?title=Special:Imagelist&sort=byname&limit=50&wpIlMatch=
    # (<a href="/en/Image:109_0923.JPG" title="Image:109 0923.JPG">desc</a>) <a href="/en/upload/c/cd/109_0923.JPG">109 0923.JPG</a> . . 885,713 bytes . . <a href="/en/User:Bfalconer" title="User:Bfalconer">Bfalconer</a> . . 18:44, 17 November 2005<br />
    ,'(?ism)<a href=[^>]+ title="[^:>]+:(?P<filename>[^>]+)">[^<]+</a>[^<]+<a href="(?P<url>[^>]+)">[^<]+</a>[^<]+<a[^>]+>(?P<uploader>[^<]+)</a>'

    # [4]
    ,(
        r'(?im)<td class="TablePager_col_img_name">\s*<a href[^>]*?>(?P<filename>[^>]+)</a>[^<]*?<a href="(?P<url>[^>]+)">[^<]*?</a>[^<]*?</td>\s*'
        r'<td class="TablePager_col_thumb">[^\n\r]*?</td>\s*'
        r'<td class="TablePager_col_img_size">[^<]*?</td>\s*'
        r'<td class="(?:TablePager_col_img_user_text|TablePager_col_img_actor)">\s*(<a href="[^>]*?" title="[^>]*?">)?(?P<uploader>[^<]+?)(</a>)?\s*</td>'
    )
]



import re

import requests

from wikiteam3.dumpgenerator.cli import Delay
from wikiteam3.dumpgenerator.api import get_JSON
from wikiteam3.dumpgenerator.config import Config

def getNamespacesScraper(config: Config, session: requests.Session):
    """Hackishly gets the list of namespaces names and ids from the dropdown in the HTML of Special:AllPages"""
    """Function called if no API is available"""
    namespaces = config.namespaces
    namespacenames = {0: ""}  # main is 0, no prefix
    if namespaces:
        r = session.post(
            url=config.index, params={"title": "Special:Allpages"}, timeout=30
        )
        raw = r.text
        Delay(config=config)

        # [^>]*? to include selected="selected"
        m = re.compile(
            r'<option [^>]*?value=[\'"](?P<namespaceid>\d+)[\'"][^>]*?>(?P<namespacename>[^<]+)</option>'
        ).finditer(raw)
        if "all" in namespaces:
            namespaces = []
            for i in m:
                namespaces.append(int(i.group("namespaceid")))
                namespacenames[int(i.group("namespaceid"))] = i.group("namespacename")
        else:
            # check if those namespaces really exist in this wiki
            namespaces2 = []
            for i in m:
                if int(i.group("namespaceid")) in namespaces:
                    namespaces2.append(int(i.group("namespaceid")))
                    namespacenames[int(i.group("namespaceid"))] = i.group(
                        "namespacename"
                    )
            namespaces = namespaces2
    else:
        namespaces = [0]

    namespaces = list(set(namespaces))  # uniques
    print("%d namespaces found" % (len(namespaces)))
    return namespaces, namespacenames


def getNamespacesAPI(config: Config, session: requests.Session):
    """Uses the API to get the list of namespaces names and ids"""
    namespaces = config.namespaces
    namespace_names = {0: ""}  # main is 0, no prefix
    if namespaces:
        r = session.get(
            url=config.api,
            params={
                "action": "query",
                "meta": "siteinfo",
                "siprop": "namespaces",
                "format": "json",
            },
            timeout=30,
        )
        result = get_JSON(r)
        Delay(config=config)
        try:
            nsquery = result["query"]["namespaces"]
        except KeyError:
            print("Error: could not get namespaces from the API request.")
            print("HTTP %d" % r.status_code)
            print(r.text)
            raise

        if "all" in namespaces:
            namespaces = []
            for i in nsquery.keys():
                if int(i) < 0:  # -1: Special, -2: Media, excluding
                    continue
                namespaces.append(int(i))
                namespace_names[int(i)] = nsquery[i]["*"]
        else:
            # check if those namespaces really exist in this wiki
            namespaces2 = []
            for i in nsquery.keys():
                bi = i
                i = int(i)
                if i < 0:  # -1: Special, -2: Media, excluding
                    continue
                if i in namespaces:
                    namespaces2.append(i)
                    namespace_names[i] = nsquery[bi]["*"]
            namespaces = namespaces2
    else:
        namespaces = [0]

    namespaces = list(set(namespaces))  # uniques
    print("%d namespaces found" % (len(namespaces)))
    return namespaces, namespace_names



from datetime import datetime
import sys
import time
from typing import Dict, List
from urllib.parse import urlparse
import lxml.etree

import mwclient
import mwclient.errors
import requests

from wikiteam3.dumpgenerator.exceptions import PageMissingError
from wikiteam3.dumpgenerator.log import log_error
from wikiteam3.dumpgenerator.api.namespaces import getNamespacesAPI
from wikiteam3.dumpgenerator.api.page_titles import read_titles
from wikiteam3.dumpgenerator.dump.page.xmlrev.xml_revisions_page import \
    make_xml_from_page, make_xml_page_from_raw
from wikiteam3.dumpgenerator.config import Config

ALL_NAMESPACE = -1

def getXMLRevisionsByAllRevisions(config: Config, session: requests.Session, site: mwclient.Site, nscontinue=None, arvcontinue=None):
    if "all" not in config.namespaces:
        namespaces = config.namespaces
    else:
        # namespaces, namespacenames = getNamespacesAPI(config=config, session=session)
        namespaces = [ALL_NAMESPACE] # magic number refers to "all"
    _nscontinue_input = nscontinue
    _arvcontinue_input = arvcontinue
    del nscontinue
    del arvcontinue

    for namespace in namespaces:
        # Skip retrived namespace
        if namespace == ALL_NAMESPACE:
            assert len(namespaces) == 1, \
                "Only one item shoule be there when 'all' namespace are specified"
            _nscontinue_input = None
        else:
            if _nscontinue_input is not None:
                if namespace != _nscontinue_input:
                    print("Skipping already exported namespace: %d" % namespace)
                    continue
                _nscontinue_input = None

        print("Trying to export all revisions from namespace %s" % namespace)
        # arvgeneratexml exists but was deprecated in 1.26 (while arv is from 1.27?!)
        arv_params = {
            "action": "query",
            "list": "allrevisions",
            "arvlimit": config.api_chunksize,
            "arvdir": "newer",
        }
        if namespace != ALL_NAMESPACE:
            arv_params['arvnamespace'] = namespace
        if _arvcontinue_input is not None:
            arv_params['arvcontinue'] = _arvcontinue_input

        if not config.curonly:
            # We have to build the XML manually...
            # Skip flags, presumably needed to add <minor/> which is in the schema.
            # Also missing: parentid and contentformat.
            arv_params[
                "arvprop"
            ] = "ids|timestamp|user|userid|size|sha1|contentmodel|comment|content|flags"
            print(
                "Trying to get wikitext from the allrevisions API and to build the XML"
            )
            while True:
                print("[arvcontinue]:", arv_params.get("arvcontinue", ""))
                try:
                    allrevs_response = site.api(
                        http_method=config.http_method, **arv_params
                    )
                except requests.exceptions.HTTPError as e:
                    if (
                            e.response.status_code == 405
                            and config.http_method == "POST"
                    ):
                        print("POST request to the API failed, retrying with GET")
                        config.http_method = "GET"
                        continue
                    else:
                        raise
                except requests.exceptions.ReadTimeout as err:
                    # Hopefully temporary, just wait a bit and continue with the same request.
                    # No point putting a limit to retries, we'd need to abort everything.
                    # TODO: reuse the retry logic of the checkAPI phase? Or force mwclient
                    # to use the retry adapter we use for our own requests session?
                    print(f"ERROR: {str(err)}")
                    print("Sleeping for 20 seconds")
                    time.sleep(20)
                    continue
                except mwclient.errors.InvalidResponse as e:
                    if (
                        e.response_text.startswith("<!DOCTYPE html>") # type: ignore
                        and config.http_method == "POST"
                    ):
                        print("POST request to the API failed (got HTML), retrying with GET")
                        config.http_method = "GET"
                        continue
                    else:
                        raise

                for page in allrevs_response["query"]["allrevisions"]:
                    yield make_xml_from_page(page, arv_params.get("arvcontinue", ""))

                # find the continue parameter
                if "continue" in allrevs_response:
                    # handle infinite loop
                    if arv_params.get("arvcontinue", None) == allrevs_response["continue"]["arvcontinue"]:
                        allrevs_response = handle_infinite_loop(
                            allrevs_response=allrevs_response, arv_params=arv_params, config=config, site=site
                        )
                    # update continue parameter
                    arv_params["arvcontinue"] = allrevs_response["continue"]["arvcontinue"]
                else:
                    # End of continuation. We are done with this namespace.
                    break

        else: # curonly
            # FIXME: this is not curonly, just different strategy to do all revisions
            # Just cycle through revision IDs and use the XML as is
            print("Trying to list the revisions and to export them one by one")
            # We only need the revision ID, all the rest will come from the raw export
            arv_params["arvprop"] = "ids"
            try:
                allrevs_response = site.api(
                    http_method=config.http_method, **arv_params
                )
            except requests.exceptions.HTTPError as e:
                if (
                        e.response.status_code == 405
                        and config.http_method == "POST"
                ):
                    print("POST request to the API failed, retrying with GET")
                    config.http_method = "GET"
                    raise NotImplementedError("FIXME: here we should retry the same namespace")
                    continue # FIXME: here we should retry the same namespace
                else:
                    raise
            export_params = {
                "action": "query",
                "export": "1",
            }
            # Skip the namespace if it's empty
            if len(allrevs_response["query"]["allrevisions"]) < 1:
                # TODO: log this
                continue
            # Repeat the arvrequest with new arvparams until done
            while True:
                # Reset revision IDs from the previous batch from arv
                revids: List[str] = []
                for page in allrevs_response["query"]["allrevisions"]:
                    for revision in page["revisions"]:
                        revids.append(str(revision["revid"]))
                print(
                    "        %d more revisions listed, until %s"
                    % (len(revids), revids[-1])
                )

                # We can now get the XML for one revision at a time
                # FIXME: we can actually get them in batches as we used to
                # but need to figure out the continuation and avoid that the API
                # chooses to give us only the latest for each page
                for revid in revids:
                    export_params["revids"] = revid
                    try:
                        export_response = site.api(
                            http_method=config.http_method, **export_params
                        )
                    except requests.exceptions.HTTPError as e:
                        if (
                                e.response.status_code == 405
                                and config.http_method == "POST"
                        ):
                            print(
                                "POST request to the API failed, retrying with GET"
                            )
                            config.http_method = "GET"
                            export_response = site.api(
                                http_method=config.http_method, **export_params
                            )
                        else:
                            raise

                    # This gives us a self-standing <mediawiki> element
                    # but we only need the inner <page>: we can live with
                    # duplication and non-ordering of page titles, but the
                    # repeated header is confusing and would not even be valid
                    xml: str = export_response["query"]["export"]["*"]
                    yield make_xml_page_from_raw(xml, arv_params.get("arvcontinue", ""))

                if "continue" in allrevs_response:
                    # Get the new ones
                    # NOTE: don't need to handle infinite loop here, because we are only getting the revids

                    arv_params["arvcontinue"] = allrevs_response["continue"]["arvcontinue"]
                    try:
                        allrevs_response = site.api(
                            http_method=config.http_method, **arv_params
                        )
                    except requests.exceptions.HTTPError as e:
                        if (
                                e.response.status_code == 405
                                and config.http_method == "POST"
                        ):
                            print(
                                "POST request to the API failed, retrying with GET"
                            )
                            config.http_method = "GET"
                            allrevs_response = site.api(
                                http_method=config.http_method, **arv_params
                            )
                    except requests.exceptions.ReadTimeout as err:
                        # As above
                        print(f"ERROR: {str(err)}")
                        print("Sleeping for 20 seconds")
                        time.sleep(20)
                        # But avoid rewriting the same revisions
                        allrevs_response["query"]["allrevisions"] = []
                        continue
                else:
                    # End of continuation. We are done with this namespace.
                    break


def getXMLRevisionsByTitles(config: Config, session: requests.Session, site: mwclient.Site, start=None):
    if config.curonly:
        # The raw XML export in the API gets a title and gives the latest revision.
        # We could also use the allpages API as generator but let's be consistent.
        print("Getting titles to export the latest revision for each")
        c = 0
        for title in read_titles(config, session=session, start=start):
            # TODO: respect verbose flag, reuse output from getXMLPage
            print(f"    {title}")
            # TODO: as we're doing one page and revision at a time, we might
            # as well use xml format and exportnowrap=1 to use the string of,
            # XML as is, but need to check how well the library handles it.
            exportparams = {
                "action": "query",
                "titles": title,
                "export": "1",
            }
            try:
                export_response = site.api(
                    http_method=config.http_method, **exportparams
                )
            except requests.exceptions.HTTPError as e:
                if (
                        e.response.status_code == 405
                        and config.http_method == "POST"
                ):
                    print("POST request to the API failed, retrying with GET")
                    config.http_method = "GET"
                    export_response = site.api(
                        http_method=config.http_method, **exportparams
                    )
                else:
                    raise

            xml = str(export_response["query"]["export"]["*"])
            c += 1
            if c % 10 == 0:
                print(f"\n->  Downloaded {c} pages\n")
            # Because we got the fancy XML from the JSON format, clean it:
            yield make_xml_page_from_raw(xml, None)
    else:
        # This is the closest to what we usually do with Special:Export:
        # take one title at a time and try to get all revisions exported.
        # It differs from the allrevisions method because it actually needs
        # to be input the page titles; otherwise, the requests are similar.
        # The XML needs to be made manually because the export=1 option
        # refuses to return an arbitrary number of revisions (see above).
        print("Getting titles to export all the revisions of each")
        c = 0
        titlelist = []
        # TODO: Decide a suitable number of a batched request. Careful:
        # batched responses may not return all revisions.
        for titlelist in read_titles(config, session=session, start=start):
            if isinstance(titlelist, str):
                titlelist = [titlelist]
            for title in titlelist:
                print(f"    {title}")
            # Try and ask everything. At least on MediaWiki 1.16, uknown props are discarded:
            # "warnings":{"revisions":{"*":"Unrecognized values for parameter 'rvprop': userid, sha1, contentmodel"}}}
            pparams = {
                "action": "query",
                "titles": "|".join(titlelist),
                "prop": "revisions",
                'rvlimit': config.api_chunksize,
                "rvprop": "ids|timestamp|user|userid|size|sha1|contentmodel|comment|content|flags",
            }
            try:
                api_response = site.api(http_method=config.http_method, **pparams)
            except requests.exceptions.HTTPError as e:
                if (
                        e.response.status_code == 405
                        and config.http_method == "POST"
                ):
                    print("POST request to the API failed, retrying with GET")
                    config.http_method = "GET"
                    api_response = site.api(
                        http_method=config.http_method, **pparams
                    )
                else:
                    raise
            except mwclient.errors.InvalidResponse:
                log_error(
                    config=config, to_stdout=True,
                    text="Error: page inaccessible? Could not export page: %s"
                         % ("; ".join(titlelist)),
                )
                continue

            # Be ready to iterate if there is continuation.
            while True:
                # Get the revision data returned by the API: prequest is the initial request
                # or the new one after continuation at the bottom of this while loop.
                # The array is called "pages" even if there's only one.
                try:
                    pages = api_response["query"]["pages"]
                except KeyError:
                    log_error(
                        config=config, to_stdout=True,
                        text="Error: page inaccessible? Could not export page: %s"
                             % ("; ".join(titlelist)),
                    )
                    break
                # Go through the data we got to build the XML.
                for pageid in pages:
                    try:
                        xml = make_xml_from_page(pages[pageid], None)
                        yield xml
                    except PageMissingError:
                        log_error(
                            config=config, to_stdout=True,
                            text="Error: empty revision from API. Could not export page: %s"
                                 % ("; ".join(titlelist)),
                        )
                        continue

                # Get next batch of revisions if there's more.
                if "continue" in api_response.keys():
                    print("Getting more revisions for the page")
                    for key, value in api_response["continue"].items():
                        pparams[key] = value
                elif "query-continue" in api_response.keys():
                    rvstartid = api_response["query-continue"]["revisions"]["rvstartid"]
                    pparams["rvstartid"] = rvstartid
                else:
                    break

                try:
                    api_response = site.api(
                        http_method=config.http_method, **pparams
                    )
                except requests.exceptions.HTTPError as e:
                    if (
                            e.response.status_code == 405
                            and config.http_method == "POST"
                    ):
                        print("POST request to the API failed, retrying with GET")
                        config.http_method = "GET"
                        api_response = site.api(
                            http_method=config.http_method, **pparams
                        )

            # We're done iterating for this title or titles.
            c += len(titlelist)
            # Reset for the next batch.
            titlelist = []
            if c % 10 == 0:
                print(f"\n->  Downloaded {c} pages\n")


def getXMLRevisions(config: Config, session: requests.Session, lastPage=None, useAllrevision=True):
    # FIXME: actually figure out the various strategies for each MediaWiki version
    apiurl = urlparse(config.api)
    # FIXME: force the protocol we asked for! Or don't verify SSL if we asked HTTP?
    # https://github.com/WikiTeam/wikiteam/issues/358
    site = mwclient.Site(
        apiurl.netloc, apiurl.path.replace("api.php", ""), scheme=apiurl.scheme, pool=session
    )

    if useAllrevision:
        # Find last title
        if lastPage is not None:
            try:
                lastNs = int(lastPage.find('ns').text)
                if False:
                    lastRevision = lastPage.find('revision')
                    lastTimestamp = lastRevision.find('timestamp').text
                    lastRevid = int(lastRevision.find('id').text)
                    lastDatetime = datetime.fromisoformat(lastTimestamp.rstrip('Z'))
                    lastArvcontinue = lastDatetime.strftime("%Y%m%d%H%M%S") + '|' + str(lastRevid)
                else:
                    lastArvcontinue = lastPage.attrib['arvcontinue']
            except Exception:
                print("Failed to find title in last trunk XML: %s" % (lxml.etree.tostring(lastPage)))
                raise
            nscontinue = lastNs
            arvcontinue = lastArvcontinue
            if not arvcontinue:
                arvcontinue = None
        else:
            nscontinue = None
            arvcontinue = None

        try:
            return getXMLRevisionsByAllRevisions(config, session, site, nscontinue, arvcontinue)
        except (KeyError, mwclient.errors.InvalidResponse) as e:
            print(e)
            # TODO: check whether the KeyError was really for a missing arv API
            print("Warning. Could not use allrevisions. Wiki too old? Try to use --xmlrevisions_page")
            sys.exit(1)
    else:
        # Find last title
        if lastPage is not None:
            try:
                start = lastPage.find('title')
            except Exception:
                print("Failed to find title in last trunk XML: %s" % (lxml.etree.tostring(lastPage)))
                raise
        else:
            start = None

        try:
            # # Uncomment these lines to raise an KeyError for testing
            # raise KeyError(999999)
            # # DO NOT UNCOMMMENT IN RELEASE
            return getXMLRevisionsByTitles(config, session, site, start)
        except mwclient.errors.MwClientError as e:
            print(e)
            print("This mwclient version seems not to work for us. Exiting.")
            sys.exit(1)


def handle_infinite_loop(allrevs_response: Dict, arv_params: Dict, config: Config, site: mwclient.Site) -> Dict:
    """
    return new allrevs_response without arvprop=content|comment if the response is truncated
    """

    assert len(allrevs_response["query"]["allrevisions"]) == 0, \
        "We should have received no revisions if we are stuck in a infinite loop"
    print("WARNING: API returned continue parameter that doesn't change, we might be stuck in a loop")
    print(f"current continue parameter: {arv_params.get('arvcontinue')}")
    print(f"API warnings: {allrevs_response.get('warnings', {})}")

    if "truncated" in allrevs_response.get("warnings",{}).get("result",{}).get("*",""):
        # workaround for [truncated API requests for "allrevisions" causes infinite loop ]
        # (https://github.com/mediawiki-client-tools/mediawiki-scraper/issues/166)
        print("Let's try to skip this revision and continue...")
        _arv_params_temp = arv_params.copy()
        # make sure response is small
        _arv_params_temp['arvprop'] = _arv_params_temp['arvprop'].replace('|content', '').replace('|comment', '') 
        _arv_params_temp["arvlimit"] = 1

        allrevs_response_new = site.api(
            http_method=config.http_method, **_arv_params_temp
        )
        assert len(allrevs_response_new["query"]["allrevisions"]) == 1, \
            "Couldn't get a single revision to skip the infinite loop" # arvlimit=1
        assert arv_params.get("arvcontinue", None) != allrevs_response_new.get("continue", {}).get("arvcontinue", None), \
            "??? Infinite loop is still there ???"
        # success, let's continue
        log_error(config=config, to_stdout=True,
                text=f"ERROR: API returned continue parameter '{arv_params.get('arvcontinue')}' that doesn't change, "
                f"skipped this revision to avoid infinite loop")
        return allrevs_response_new
    else:
        raise NotImplementedError("Unable to solve the infinite loop automatically")


import datetime
import os
import random
import re
import shutil
import sys
import time
import urllib.parse
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import requests

from wikiteam3.dumpgenerator.api import get_JSON, handle_StatusCode
from wikiteam3.dumpgenerator.cli import Delay
from wikiteam3.dumpgenerator.config import Config
from wikiteam3.dumpgenerator.dump.image.html_regexs import R_NEXT, REGEX_CANDIDATES
from wikiteam3.dumpgenerator.exceptions import FileSha1Error, FileSizeError
from wikiteam3.dumpgenerator.log import log_error
from wikiteam3.dumpgenerator.version import getVersion
from wikiteam3.utils.identifier import url2prefix_from_config
from wikiteam3.utils.monkey_patch import SessionMonkeyPatch
from wikiteam3.utils.util import clean_HTML, int_or_zero, sha1bytes, sha1sum, space, underscore, undo_HTML_entities

NULL = "null"
""" NULL value for image metadata"""


WBM_EARLIEST = 1
WBN_LATEST = 2
WBM_BEST = 3


def check_response(r: requests.Response) -> None:
    if r.headers.get("cf-polished", ""):
        raise RuntimeError("Found cf-polished header in response, use --bypass-cdn-image-compression to bypass it")

class Image:

    @staticmethod
    def generate_image_dump(config: Config, other: Dict, images: List[List],
                            session: requests.Session):
        """ Save files and descriptions using a file list """

        bypass_cdn_image_compression: bool = other["bypass_cdn_image_compression"]
        disable_image_verify: bool = other["disable_image_verify"]
        image_timestamp_interval: Optional[str] = other["image_timestamp_interval"]
        ia_wbm_booster: int = other["ia_wbm_booster"]
        add_referer_header: Optional[str] = other["add_referer_header"] # None, "auto", {URL}

        image_timestamp_intervals = None
        if image_timestamp_interval: # 2019-01-02T01:36:06Z/2023-08-12T10:36:06Z
            image_timestamp_intervals = image_timestamp_interval.split("/")
            assert len(image_timestamp_intervals) == 2
            image_timestamp_intervals = [
                datetime.datetime.strptime(x, "%Y-%m-%dT%H:%M:%SZ")
                for x in image_timestamp_intervals]

        print("Retrieving images...")
        images_dir = Path(config.path) / "images"
        if not os.path.isdir(images_dir):
            print(f'Creating "{images_dir}" directory')
            os.makedirs(images_dir)

        c_savedImageFiles = 0
        c_wbm_speedup_files = 0


        def modify_params(params: Optional[Dict] = None) -> Dict:
            """ bypass Cloudflare Polish (image optimization) """
            if params is None:
                params = {}
            if bypass_cdn_image_compression is True:
                # bypass Cloudflare Polish (image optimization)
                # <https://developers.cloudflare.com/images/polish/>
                params["_wiki_t"] = int(time.time()*1000)
                params[f"_wiki_{random.randint(10,99)}_"] = "random"

            return params
        
        def modify_headers(headers: Optional[Dict] = None) -> Dict:
            """ add HTTP Referer header """
            if headers is None:
                headers = {}
            if add_referer_header:
                url = config.index if config.index else config.api
                parsed_url = urllib.parse.urlparse(
                    add_referer_header
                    if add_referer_header != "auto"
                    else url
                )

                headers["Referer"] = f"{parsed_url.scheme}://{parsed_url.netloc}/"

            return headers
            

        patch_sess = SessionMonkeyPatch(session=session, config=config, hard_retries=3)
        patch_sess.hijack()

        ia_session = requests.Session()
        ia_session.headers.update({"User-Agent": f"wikiteam3/{getVersion()}"})

        skip_to_filename = underscore('') # TODO: use this

        while images:
            filename_raw, url_raw, uploader_raw, size, sha1, timestamp \
                = images.pop(0) # reduce memory usage by poping
            filename_underscore = underscore(filename_raw)
            # uploader_underscore = space(uploader_raw)

            if skip_to_filename and skip_to_filename != filename_underscore:
                print(f"    {filename_underscore}", end="\r")
                continue
            else:
                skip_to_filename = ''

            downloaded = False

            if image_timestamp_intervals:
                if timestamp == NULL:
                    print(f"    {filename_underscore}|timestamp is unknown: {NULL}, downloading anyway...")
                else:
                    if not (
                        image_timestamp_intervals[0]
                        <= datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ")
                        <= image_timestamp_intervals[1]
                    ):
                        print(f"    timestamp {timestamp} is not in interval {image_timestamp_interval}: {filename_underscore}")
                        continue
                    else:
                        print(f"    timestamp {timestamp} is in interval {image_timestamp_interval}: {filename_underscore}")

            # saving file
            if filename_underscore != urllib.parse.unquote(filename_underscore):
                print(f"WARNING:    {filename_underscore}|filename may not be unquoted: {filename_underscore}")
            if len(filename_underscore.encode('utf-8')) > other["filenamelimit"]:
                log_error(
                    config=config, to_stdout=True,
                    text=f"Filename is too long(>{other['filenamelimit']} bytes), skipping: '{filename_underscore}'",
                )
                # TODO: hash as filename instead of skipping
                continue

            filepath_space = images_dir / space(filename_underscore)
            filepath_underscore = images_dir / filename_underscore

            if filepath_space.is_file():
                # rename file to underscore
                shutil.move(filepath_space, filepath_underscore)

            # check if file already exists and has the same size and sha1
            if ((size != NULL
                and filepath_underscore.is_file()
                and os.path.getsize(filepath_underscore) == int(size)
                and sha1sum(filepath_underscore) == sha1)
            or (sha1 == NULL and filepath_underscore.is_file())): 
            # sha1 is NULL if file not in original wiki (probably deleted,
            # you will get a 404 error if you try to download it)
                c_savedImageFiles += 1
                downloaded = True
                print_msg=f"    {c_savedImageFiles}|sha1 matched: {filename_underscore}"
                print(print_msg[0:70], end="\r")
                if sha1 == NULL:
                    log_error(config=config, to_stdout=True,
                        text=f"sha1 is {NULL} for {filename_underscore}, file may not in wiki site (probably deleted). "
                    )
            else:
                # Delay(config=config, delay=config.delay + random.uniform(0, 1))
                url = url_raw

                r: Optional[requests.Response] = None
                if ia_wbm_booster:
                    def get_ia_wbm_response() -> Optional[requests.Response]:
                        """ Get response from Internet Archive Wayback Machine
                        return None if not found / failed """
                        if ia_wbm_booster in (WBM_EARLIEST, WBN_LATEST):
                            ia_timestamp = ia_wbm_booster
                        elif ia_wbm_booster == WBM_BEST:
                            if timestamp != NULL:
                                ia_timestamp = [x for x in timestamp if x.isdigit()][0:8]
                                ia_timestamp = "".join(ia_timestamp)
                            else:
                                print(f"ia_wbm_booster:    timestamp is {NULL}, use latest timestamp")
                                ia_timestamp = 2
                        else:
                            raise ValueError(f"ia_wbm_booster is {ia_wbm_booster}, but it should be 0, 1, 2 or 3")

                        available_api = "http://archive.org/wayback/available"
                        # TODO: cdx_api = "http://web.archive.org/cdx/search/cdx"
                        snap_url = f"https://web.archive.org/web/{ia_timestamp}id_/{url}"

                        try:
                            _r = ia_session.get(available_api, params={"url": url}, headers={"User-Agent": "wikiteam3"},
                                            timeout=10)
                            if _r.status_code == 429:
                                raise Warning("IA API rate limit exceeded (HTTP 429)")
                            _r.raise_for_status()
                            api_result = _r.json()
                            if api_result["archived_snapshots"]:
                                r = ia_session.get(url=snap_url, allow_redirects=True)
                                # r.raise_for_status()
                            else:
                                r = None
                        except Exception as e:
                            print("ia_wbm_booster:",e)
                            r = None

                        return r
                    r = get_ia_wbm_response()

                    # verify response
                    if r and r.status_code != 200:
                        r = None
                    elif r and size != NULL and len(r.content) != int(size): # and r.status_code == 200:
                        # FileSizeError
                        # print(f"WARNING:    {filename_unquoted} size should be {size}, but got {len(r.content)} from WBM, use original url...")
                        r = None

                    if r is not None:
                        c_wbm_speedup_files += 1


                if r is None:
                    Delay(config=config)
                    r = session.get(url=url, params=modify_params(), headers=modify_headers(), allow_redirects=True)
                    check_response(r)

                    # a trick to get original file (fandom)
                    ori_url = url
                    if "fandom.com" in config.api \
                        and "static.wikia.nocookie.net" in url \
                        and "?" in url \
                        and (
                               sha1 != NULL and sha1bytes(r.content) != sha1
                            or size != NULL and len(r.content) != int(size)
                        ):
                        ori_url = url + "&format=original"
                        Delay(config=config)
                        r = session.get(url=ori_url, params=modify_params(), headers=modify_headers(), allow_redirects=True)
                        check_response(r)

                    # Try to fix a broken HTTP to HTTPS redirect
                    original_url_redirected: bool = r.url in (url, ori_url)
                    if r.status_code == 404 and original_url_redirected:
                        print(f"WARNING: {url} broken (404), trying to fix it...")
                        if (
                            url_raw.startswith("http://")
                            and url.startswith("https://")
                        ):
                            url = "https://" + url_raw.split("://")[1]
                            # print 'Maybe a broken http to https redirect, trying ', url
                            r = session.get(url=url, params=modify_params(), headers=modify_headers(), allow_redirects=True)
                            check_response(r)

                if r.status_code == 200:
                    try:
                        if disable_image_verify \
                            or (sha1 == NULL and size == NULL) \
                            or (
                                    (sha1 == NULL or sha1bytes(r.content) == sha1)
                                and (size == NULL or len(r.content) == int(size) )
                            ):
                            try:
                                with open(filepath_underscore, "wb") as imagefile:
                                    imagefile.write(r.content)
                            except KeyboardInterrupt:
                                if filepath_underscore.is_file():
                                    os.remove(filepath_underscore)
                                raise
                            c_savedImageFiles += 1
                        else:
                            if len(r.content) != int(size):
                                raise FileSizeError(file=filename_underscore,
                                                    got_size=len(r.content),
                                                    excpected_size=int(size),
                                                    online_url=url)
                            elif sha1bytes(r.content) != sha1:
                                raise FileSha1Error(file=filename_underscore, excpected_sha1=sha1)
                            else:
                                raise RuntimeError("Unknown error")
                    except OSError:
                        log_error(
                            config=config, to_stdout=True,
                            text=f"File '{filepath_underscore}' could not be created by OS",
                        )
                        continue
                    except FileSizeError as e:
                        log_error(
                            config=config, to_stdout=True,
                            text=f"File '{e.file}' size is not match '{e.excpected_size}', skipping",
                        )
                        continue
                    except FileSha1Error as e:
                        log_error(
                            config=config, to_stdout=True,
                            text=f"File '{e.file}' sha1 is not match '{e.excpected_sha1}', skipping",
                        )
                        continue

                    if timestamp != NULL:
                        # try to set file timestamp (mtime)
                        try:
                            mtime = datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ").timestamp()
                            atime = os.stat(filepath_underscore).st_atime
                            # atime is not modified
                            os.utime(filepath_underscore, times=(atime, mtime))
                            # print(atime, mtime)
                        except Exception as e:
                            print("Error setting file timestamp:", e)
                else:
                    log_error(
                        config=config, to_stdout=True,
                        text=f"Failed to download '{filename_underscore}' with URL '{url}' due to HTTP '{r.status_code}', skipping"
                    )

            if downloaded: # skip printing
                continue
            print_msg = f"              | {len(images)}=>{filename_underscore[0:50]}"
            print(print_msg, " "*(73 - len(print_msg)), end="\r")

        # NOTE: len(images) == 0 here

        patch_sess.release()
        print(f"Downloaded {c_savedImageFiles} files")
        if ia_wbm_booster and c_wbm_speedup_files:
            print(f"(WBM speedup: {c_wbm_speedup_files} files)")


    @staticmethod
    def get_image_names(config: Config, session: requests.Session):
        """Get list of image names"""

        print(")Retrieving image filenames")
        images = []
        if config.api:
            print("Using API to retrieve image names...")
            images = Image.get_image_names_API(config=config, session=session)
        elif config.index:
            print("Using index.php (Special:Imagelist) to retrieve image names...")
            images = Image.get_image_names_scraper(config=config, session=session)

        print(f"Sorting image filenames ({len(images)} images)...")
        images.sort()
        print("Done")

        return images


    @staticmethod
    def get_image_names_scraper(config: Config, session: requests.Session):
        """Retrieve file list: filename, url, uploader"""

        images = []
        limit = 5000
        retries = config.retries
        offset = None
        while offset or len(images) == 0:
            # 5000 overload some servers, but it is needed for sites like this with
            # no next links
            # http://www.memoryarchive.org/en/index.php?title=Special:Imagelist&sort=byname&limit=50&wpIlMatch=
            params = {"title": "Special:Imagelist", "limit": limit, "dir": "prev", "offset": offset}
            r = session.post(
                url=config.index,
                params=params,
                timeout=30,
            )
            raw = r.text
            Delay(config=config)
            # delicate wiki
            if re.search(
                r"(?i)(allowed memory size of \d+ bytes exhausted|Call to a member function getURL)",
                raw,
            ):
                if limit > 10:
                    print(f"Error: listing {limit} images in a chunk is not possible, trying tiny chunks")
                    limit = limit // 10
                    continue
                elif retries > 0:  # waste retries, then exit
                    retries -= 1
                    print("Retrying...")
                    continue
                else:
                    raise RuntimeError("retries exhausted")

            raw = clean_HTML(raw)

            # Select the regexp that returns more results
            best_matched = 0
            regexp_best = None
            for regexp in REGEX_CANDIDATES:
                _count = len(re.findall(regexp, raw))
                if _count > best_matched:
                    best_matched = _count
                    regexp_best = regexp
            assert regexp_best is not None, "Could not find a proper regexp to parse the HTML"
            m = re.compile(regexp_best).finditer(raw)

            # Iter the image results
            for i in m:
                url = i.group("url")
                url = Image.curate_image_URL(config=config, url=url)

                filename = i.group("filename")
                filename = undo_HTML_entities(text=filename)
                filename = urllib.parse.unquote(filename)

                uploader = i.group("uploader")
                uploader = undo_HTML_entities(text=uploader)
                uploader = urllib.parse.unquote(uploader)

                # timestamp = i.group("timestamp")
                # print("    %s" % (timestamp))

                size = NULL # size not accurate
                sha1 = NULL # sha1 not available
                timestamp = NULL # date formats are difficult to parse
                images.append([
                    underscore(filename), url, space(uploader),
                    size, sha1, timestamp,
                ])
                # print (filename, url)

            if re.search(R_NEXT, raw):
                new_offset = re.findall(R_NEXT, raw)[0]
                # Avoid infinite loop
                if new_offset != offset:
                    offset = new_offset
                    retries += 5  # add more retries if we got a page with offset
                else:
                    print("Warning: offset is not changing")
                    offset = ""
            else:
                print("INFO: no next link found, we may have reached the end")
                offset = ""

        if len(images) == 0:
            print("Warning: no images found")
        elif len(images) == limit:
            print(f"Warning: the number of images is equal to the limit parameter ({limit}), there may be more images")
        else:
            print(f"    Found {len(images)} images")

        images.sort()
        return images

    @staticmethod
    def get_image_names_API(config: Config, session: requests.Session):
        """Retrieve file list: filename, url, uploader, size, sha1"""
        use_oldAPI = False
        # # Commented by @yzqzss:
        # https://www.mediawiki.org/wiki/API:Allpages
        # API:Allpages requires MW >= 1.8
        # API:Allimages requires MW >= 1.13

        aifrom = "!"
        images = []
        countImages = 0
        while aifrom:
            print(f'Using API:Allimages to get the list of images, {len(images)} images found so far...', end='\r')
            params = {
                "action": "query",
                "list": "allimages",
                "aiprop": "url|user|size|sha1|timestamp",
                "aifrom": aifrom,
                "format": "json",
                "ailimit": config.api_chunksize,
            }
            # FIXME Handle HTTP Errors HERE
            r = session.get(url=config.api, params=params, timeout=30)
            handle_StatusCode(r)
            jsonimages = get_JSON(r)
            Delay(config=config)

            if "query" in jsonimages:
                countImages += len(jsonimages["query"]["allimages"])
                
                # oldAPI = True
                # break
                # # uncomment to force use API:Allpages generator 
                # # may also can as a fallback if API:Allimages response is wrong

                aifrom = ""
                if (
                    "query-continue" in jsonimages
                    and "allimages" in jsonimages["query-continue"]
                ):
                    if "aicontinue" in jsonimages["query-continue"]["allimages"]:
                        aifrom = jsonimages["query-continue"]["allimages"]["aicontinue"]
                    elif "aifrom" in jsonimages["query-continue"]["allimages"]:
                        aifrom = jsonimages["query-continue"]["allimages"]["aifrom"]
                elif "continue" in jsonimages:
                    if "aicontinue" in jsonimages["continue"]:
                        aifrom = jsonimages["continue"]["aicontinue"]
                    elif "aifrom" in jsonimages["continue"]:
                        aifrom = jsonimages["continue"]["aifrom"]
                print(countImages, aifrom[0:30]+" "*(60-len(aifrom[0:30])),end="\r")

                for image in jsonimages["query"]["allimages"]:
                    image: Dict

                    url = image["url"]
                    url = Image.curate_image_URL(config=config, url=url)

                    filename = image.get("name", None)
                    if filename is None:
                        if  (
                            ".wikia." in config.api or ".fandom.com" in config.api
                        ):
                            filename = urllib.parse.unquote(
                                url.split("/")[-3]
                            )
                        else:
                            filename = urllib.parse.unquote(
                                url.split("/")[-1]
                            )
                    assert " " not in filename, f'Filename "{filename}" contains space, please report this to the developer(s)'

                    if "%u" in filename:
                        warnings.warn(
                            f"Filename {filename} may contains unquoted URL characters, please review it manually. FILENAME: {filename} URL:{url}",
                            UnicodeWarning,
                        )

                    uploader = image.get("user", "Unknown")
                    size: Union[bool,int] = image.get("size", NULL)
                    
                    # size or sha1 is not always available (e.g. https://wiki.mozilla.org/index.php?curid=20675)
                    sha1: Union[bool,str] = image.get("sha1", NULL)
                    timestamp = image.get("timestamp", NULL)
                    images.append([underscore(filename), url, space(uploader), size, sha1, timestamp])
            else:
                use_oldAPI = True
                break

        if use_oldAPI:
            print("    API:Allimages not available. Using API:Allpages generator instead.")
            gapfrom = "!"
            images = []
            while gapfrom:
                # Some old APIs doesn't have allimages query
                # In this case use allpages (in nm=6) as generator for imageinfo
                # Example:
                # http://minlingo.wiki-site.com/api.php?action=query&generator=allpages&gapnamespace=6
                # &gaplimit=500&prop=imageinfo&iiprop=user|url&gapfrom=!
                params = {
                    "action": "query",
                    "generator": "allpages",
                    "gapnamespace": 6,
                    "gaplimit": config.api_chunksize, # The value must be between 1 and 500.
                                    # TODO: Is it OK to set it higher, for speed?
                    "gapfrom": gapfrom,
                    "prop": "imageinfo",
                    "iiprop": "url|user|size|sha1|timestamp",
                    "format": "json",
                }
                # FIXME Handle HTTP Errors HERE
                r = session.get(url=config.api, params=params, timeout=30)
                handle_StatusCode(r)
                jsonimages = get_JSON(r)
                Delay(config=config)

                if "query" in jsonimages:
                    countImages += len(jsonimages["query"]["pages"])
                    print(countImages, gapfrom[0:30]+" "*(60-len(gapfrom[0:30])),end="\r")

                    gapfrom = ""

                    # all moden(at 20221231) wikis return 'continue' instead of 'query-continue'
                    if (
                        "continue" in jsonimages
                        and "gapcontinue" in jsonimages["continue"]
                    ):
                        gapfrom = jsonimages["continue"]["gapcontinue"]
                    
                    # legacy code, not sure if it's still needed by some old wikis
                    elif (
                        "query-continue" in jsonimages
                        and "allpages" in jsonimages["query-continue"]
                    ):
                        if "gapfrom" in jsonimages["query-continue"]["allpages"]:
                            gapfrom = jsonimages["query-continue"]["allpages"][
                                "gapfrom"
                            ]


                    # print (gapfrom)
                    # print (jsonimages['query'])

                    for image, props in jsonimages["query"]["pages"].items():
                        url = props["imageinfo"][0]["url"]
                        url = Image.curate_image_URL(config=config, url=url)

                        filename = ":".join(props["title"].split(":")[1:])

                        uploader = props["imageinfo"][0]["user"]
                        size = props.get("imageinfo")[0].get("size", NULL)
                        sha1 = props.get("imageinfo")[0].get("sha1", NULL)
                        timestamp = props.get("imageinfo")[0].get("timestamp", NULL)
                        images.append([underscore(filename), url, space(uploader), size, sha1, timestamp])
                else:
                    # if the API doesn't return query data, then we're done
                    break

        if len(images) == 1:
            print("    Found 1 image")
        else:
            print("    Found %d images" % (len(images)))

        return images


    @staticmethod
    def save_image_names(config: Config, images: List[List]):
        """Save image list in a file, including filename, url, uploader and other metadata"""

        images_filename = "{}-{}-images.txt".format(
            url2prefix_from_config(config=config), config.date
        )
        images_file = open(
            "{}/{}".format(config.path, images_filename), "w", encoding="utf-8"
        )

        c_images_size = 0
        for line in images:
            while 3 <= len(line) < 6:
                line.append(NULL) # At this point, make sure all lines have 5 elements
            filename, url, uploader, size, sha1, timestamp = line

            assert " " not in filename, "Filename contains space, it should be underscored"
            assert "_" not in uploader, "Uploader contains underscore, it should be spaced"

            # print(line,end='\r')
            c_images_size += int_or_zero(size)

            images_file.write(
                filename + "\t" + url + "\t" + uploader
                + "\t" + (str(size) if size else NULL)
                + "\t" + (str(sha1) if sha1 else NULL) # sha1 or size may be NULL
                + "\t" + (timestamp if timestamp else NULL)
                + "\n"
            )
        images_file.write("--END--\n")
        images_file.close()

        print("Image metadata (images.txt) saved at:", images_filename)
        print(f"Estimated size of all images (images.txt): {c_images_size} bytes ({c_images_size/1024/1024/1024:.2f} GiB)")


    @staticmethod
    def curate_image_URL(config: Config, url: str):
        """Returns an absolute URL for an image, adding the domain if missing"""

        if config.index:
            # remove from :// (http or https) until the first / after domain
            domainalone = (
                config.index.split("://")[0]
                + "://"
                + config.index.split("://")[1].split("/")[0]
            )
        elif  config.api:
            domainalone = (
                config.api.split("://")[0]
                + "://"
                + config.api.split("://")[1].split("/")[0]
            )
        else:
            print("ERROR: no index nor API")
            sys.exit(1)
            return # useless but linting is happy

        if url.startswith("//"):  # Orain wikifarm returns URLs starting with //
            url = "{}:{}".format(domainalone.split("://")[0], url)
        # is it a relative URL?
        elif url[0] == "/" or (
            not url.startswith("http://") and not url.startswith("https://")
        ):
            if url[0] == "/":  # slash is added later
                url = url[1:]
            # concat http(s) + domain + relative url
            url = f"{domainalone}/{url}"
        url = undo_HTML_entities(text=url)
        # url = urllib.parse.unquote(url) #do not use unquote with url, it break some
        # urls with odd chars

        return underscore(url)



import itertools
import threading
import time
from typing import Optional

from wikiteam3.dumpgenerator.config import Config, load_config
import timeit

class Delay:
    done: bool = False
    lock: threading.Lock = threading.Lock()

    def animate(self):
        progress_dots = itertools.cycle([".", "/", "-", "\\"])
        for dot in progress_dots:
            with self.lock:
                if self.done:
                    return

                print("\r" + self.ellipses, end=dot)

            time.sleep(4)

    def __init__(self, config: Optional[Config]=None,
                 msg: Optional[str]=None, delay: Optional[float]=None, dynamic: bool=True):
        """Add a delay if configured for that
        
        if delay is None, use config.delay
        if dynamic is True, load config.json every time delay is called
        """

        if delay is None:
            assert isinstance(config, Config)
            if dynamic:
                assert config
                t = timeit.default_timer()
                try:
                    config_dynamic = load_config(config=config, config_filename="config.json")
                except Exception as e:
                    print(e)
                    config_dynamic = config
                delay = config_dynamic.delay - (timeit.default_timer() - t) # compensation
            else:
                delay = config.delay
        if delay <= 0:
            return

        if msg:
            self.ellipses = ("Delay %.1fs: %s " % (delay, msg))
        else:
            self.ellipses = ("Delay %.1fs" % (delay))

        ellipses_animation = threading.Thread(target=self.animate)
        ellipses_animation.daemon = True
        ellipses_animation.start()

        time.sleep(delay)

        with self.lock:
            self.done = True
            print("\r" + " " * len(self.ellipses), end=" \r")



import re
import time

import requests

class print_probability:
    """ Decorator for printing the return value of a function """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        ret = self.func(*args, **kwargs)
        print(f"index.php available probability: {ret*100:.0f}% ({ret})")
        return ret

@print_probability
def check_index(*, index: str, logged_in: bool, session: requests.Session) -> float:
    """ Checking index.php availability
    
    returns:
        the probability of index.php being available.
        * [0.0 - 0.5) - not available
        * 0.5 - generally not sure
        * (0.5 - 1] - available
    """

    print("Checking index.php...", index)

    r = None

    for page in [
                 "Special:Random",
                 "Special:Version",
                 "Special:AllPages",
                 "Special:ListFiles",
                 "Special:Search",
                 ]:
        print(f"check_index(): Trying {page}...")

        try:
            try:
                r = session.post(url=index, data={"title": page}, timeout=30, allow_redirects=True)
            except requests.exceptions.TooManyRedirects:
                r = session.post(url=index, params={"title": page}, timeout=30, allow_redirects=False)
        except Exception as e:
            print("check_index(): Exception:", e)
            time.sleep(2)
            continue

        for _r in r.history:
            print(_r.request.method, _r.url, {"title": page}, _r.status_code)
        print(r.request.method, r.url, {"title": page}, r.status_code)

        if r.status_code in [301, 302, 303, 307, 308]:
            print("The index.php returned a redirect")
            continue

        if r.status_code >= 400:
            print(f"ERROR: The wiki returned status code HTTP {r.status_code}")
            continue

        break

    if r is None:
        print("ERROR: Failed to get index.php")
        return 0.15
    
    if r.status_code in [301, 302, 303, 307, 308]:
        print("The index.php returned a redirect")
        return 0.3

    raw = r.text
    # Workaround for
    # [Don't try to download private wikis unless --cookies is given]
    # (https://github.com/wikiTeam/wikiteam/issues/71)
    if (
        re.search(
            '(Special:Badtitle</a>|class="permissions-errors"|"wgCanonicalSpecialPageName":"Badtitle"|Login Required</h1>)',
            raw,
        )
        and not logged_in
    ):
        print("ERROR: This wiki requires login and we are not authenticated")
        return 0.5
    if re.search(
        '(page-Index_php|"wgPageName":"Index.php"|"firstHeading"><span dir="auto">Index.php</span>)',
        raw,
    ):
        print("Looks like the page called Index.php, not index.php itself")
        return 0.1
    if re.search(
        '(This wiki is powered by|<h2 id="mw-version-license">|meta name="generator" content="MediaWiki|class="mediawiki)',
        raw,
    ):
        return 0.9

    return 0.2



import argparse
from datetime import datetime
import json
import os
import random
import re
import shutil
from dataclasses import dataclass
import sys
import time
import traceback
from typing import Dict, List, Optional, Tuple, Union
import urllib.parse
from io import BytesIO
from pathlib import Path

import requests
from internetarchive import get_item, Item
from file_read_backwards import FileReadBackwards

from wikiteam3.dumpgenerator.api.page_titles import checkTitleOk
from wikiteam3.dumpgenerator.config import Config, load_config
from wikiteam3.dumpgenerator.version import getVersion
from wikiteam3.uploader.socketLock import NoLock, SocketLockServer
from wikiteam3.utils import url2prefix_from_config, sha1sum
from wikiteam3.uploader.compresser import ZstdCompressor, SevenZipCompressor
from wikiteam3.utils.ia_checker import ia_s3_tasks_load_avg
from wikiteam3.utils.util import ALL_DUMPED_MARK, UPLOADED_MARK, mark_as_done, is_markfile_exists

DEFAULT_COLLECTION = 'opensource'
TEST_COLLECTION = 'test_collection'
"""
items here are expected to be automatically removed after 30 days. 
(see <https://archive.org/details/test_collection?tab=about>)
"""
WIKITEAM_COLLECTION = 'wikiteam'
""" Only admins can add/move items to this collection. """

IDENTIFIER_PREFIX = "wiki-"

@dataclass
class IAKeys:
    access: str
    secret: str

@dataclass
class Args:
    keys_file: Path
    collection: str
    dry_run: bool
    update: bool
    wikidump_dir: Path

    bin_zstd: str
    zstd_level: int
    bin_7z: str
    parallel: bool

    def __post_init__(self):
        self.keys_file = Path(self.keys_file).expanduser().resolve()
        if not self.keys_file.exists():
            raise FileNotFoundError(f"Keys file {self.keys_file} does not exist")
        self.wikidump_dir = Path(self.wikidump_dir).expanduser().resolve()
        if not self.wikidump_dir.exists():
            raise FileNotFoundError(f"wikidump_dir {self.wikidump_dir} does not exist")


def read_ia_keys(path: Path) -> IAKeys:
    with open(path.expanduser().resolve()) as f:
        lines = f.readlines()

    access = lines[0].strip()
    secret = lines[1].strip()

    return IAKeys(access, secret)


def config2basename(config: Config) -> str:
    basename = "{}-{}".format(
        url2prefix_from_config(config=config),
        config.date,
    )
    return basename


def xmldump_is_complete(xml_path: Union[str, Path]) -> bool:
    lines_left = 100
    with FileReadBackwards(xml_path, encoding="utf-8") as frb:
        for l in frb:
            if l.strip() == "</mediawiki>":
                # xml dump is complete
                return True

            lines_left -= 1
            if lines_left <= 0:
                return False
    
    return False


def images_list_is_complete(images_txt_path: Union[str, Path]) -> bool:
    lines_left = 3
    with FileReadBackwards(images_txt_path, encoding="utf-8") as frb:
        for l in frb:
            if l.strip() == "--END--":
                # images list is complete
                return True

            lines_left -= 1
            if lines_left <= 0:
                return False
    
    return False

def get_xml_filename(config: Config) -> str:
    xml_filename = "{}-{}.xml".format(
        config2basename(config),
        "current" if config.curonly else "history",
    )
    return xml_filename


def prepare_xml_zst_file(wikidump_dir: Path, config: Config, *, parallel: bool, zstd_level: int) -> Path:
    """ Compress xml file to .zst file."""
    xml_filename = get_xml_filename(config)

    xml_file_path = wikidump_dir / xml_filename
    xml_zstd_file_path = wikidump_dir / f"{xml_filename}.zst"

    assert xml_file_path.exists() or xml_zstd_file_path.exists()

    if xml_file_path.exists():
        assert xmldump_is_complete(xml_file_path)
        with NoLock() if parallel else SocketLockServer():
            # ensure only one process is compressing, to avoid OOM
            r = ZstdCompressor.compress_file(xml_file_path, level=zstd_level)
            assert r == xml_zstd_file_path.resolve()
            assert xml_zstd_file_path.exists()
            assert ZstdCompressor.test_integrity(r)

            # rm source xml file
            # decompressing is so fast that we don't need to keep the xml file
            # os.remove(xml_file_path)

    assert xml_zstd_file_path.exists()

    return xml_zstd_file_path.resolve()


def prepare_images_7z_archive(wikidump_dir: Path, config: Config, parallel: bool) -> Path:

    images_dir = wikidump_dir / "images"
    assert images_dir.exists() and images_dir.is_dir()

    images_7z_archive_path = wikidump_dir / f"{config2basename(config)}-images.7z"
    if not images_7z_archive_path.exists() or not images_7z_archive_path.is_file():
        with NoLock() if parallel else SocketLockServer():
            r = SevenZipCompressor.compress_dir(images_dir)
            shutil.move(r, images_7z_archive_path)

    assert SevenZipCompressor.test_integrity(images_7z_archive_path)

    assert images_7z_archive_path.exists() and images_7z_archive_path.is_file()
    return images_7z_archive_path.resolve()


def prepare_files_to_upload(wikidump_dir: Path, config: Config, item: Item, *, parallel: bool, zstd_level: int) -> Dict[str, str]:
    """ return: filedict ("remote filename": "local filename") """
    filedict = {} # "remote filename": "local filename"

    # config.json
    config_json_path = wikidump_dir / "config.json"
    assert config_json_path.exists()
    filedict[f"{config2basename(config)}-dumpMeta/config.json"] = str(config_json_path)

    # errors.log optional
    if (wikidump_dir / "errors.log").exists():
        filedict[f"{config2basename(config)}-dumpMeta/errors.log"] = str(wikidump_dir / "errors.log")
    # SpecialVersion.html optional
    if (wikidump_dir / "SpecialVersion.html").exists():
        filedict[f"{config2basename(config)}-dumpMeta/SpecialVersion.html"] = str(wikidump_dir / "SpecialVersion.html")
    # siteinfo.json optional
    if (wikidump_dir / "siteinfo.json").exists():
        filedict[f"{config2basename(config)}-dumpMeta/siteinfo.json"] = str(wikidump_dir / "siteinfo.json")
    # index.html optional
    if (wikidump_dir / "index.html").exists():
        filedict[f"{config2basename(config)}-dumpMeta/index.html"] = str(wikidump_dir / "index.html")

    print("=== commpressing necessary files: ===")

    # .xml dump
    if config.xml:
        if not config.xmlrevisions:
            #  -titles.txt
            titles_txt_path = wikidump_dir / f"{config2basename(config)}-titles.txt"
            titles_txt_zstd_path = wikidump_dir / f"{config2basename(config)}-titles.txt.zst"
            assert titles_txt_path.exists()
            assert checkTitleOk(config)
            r = ZstdCompressor.compress_file(titles_txt_path,level=zstd_level)
            assert r == titles_txt_zstd_path.resolve()
            assert ZstdCompressor.test_integrity(r)
            filedict[f"{config2basename(config)}-dumpMeta/{titles_txt_zstd_path.name}"] = str(titles_txt_zstd_path)
        xml_zstd_path = prepare_xml_zst_file(wikidump_dir, config, parallel=parallel, zstd_level=zstd_level)
        filedict[f"{xml_zstd_path.name}"] = str(xml_zstd_path)

    # images
    if config.images:
        # images.txt
        images_txt_path = wikidump_dir / f"{config2basename(config)}-images.txt"
        images_txt_zstd_path = wikidump_dir / f"{config2basename(config)}-images.txt.zst"
        assert images_list_is_complete(images_txt_path)
        r = ZstdCompressor.compress_file(images_txt_path, level=zstd_level)
        assert r == images_txt_zstd_path.resolve()
        assert ZstdCompressor.test_integrity(r)
        filedict[f"{config2basename(config)}-dumpMeta/{images_txt_zstd_path.name}"] = str(images_txt_zstd_path)

        # images.7z
        images_7z_archive_path = prepare_images_7z_archive(wikidump_dir, config, parallel)
        filedict[f"{images_7z_archive_path.name}"] = str(images_7z_archive_path)

    print("=== Files already uploaded: ===")
    c = 0
    for file_in_item in item.files:
        if file_in_item["name"] in filedict:
            c += 1
            if int(file_in_item["size"]) != os.path.getsize(filedict[file_in_item["name"]]):
                print(f'    "{file_in_item["name"]}" (size mismatch), will re-upload')
                continue

            filedict.pop(file_in_item["name"])
            print(f'    "{file_in_item["name"]}" (already uploaded)')
    print(f"Already uploaded {c} files. ({len(item.files)} files in remote item in total)")

    print("=== Files to upload: ===")
    for remote_dest, local_src in filedict.items():
        print(f'    "{remote_dest}" from "{local_src}"')
    print(f"{len(filedict)} files ready to upload...")

    return filedict

def prepare_item_metadata(wikidump_dir: Path, config: Config, arg: Args) -> Tuple[Dict, Optional[str]]:
    """ return: (IA item metadata dict, logo_url) """

    wiki_prefix: str = url2prefix_from_config(config=config, ascii_slugify=False) # e.g. wiki.example.org

    sitename: Optional[str] = None # or empty str
    rights_text: Optional[str] = None # or empty str
    rights_url: Optional[str] = None # or empty str
    lang: Optional[str] = None # or empty str
    base_url: Optional[str] = None # or empty str
    logo_url: Optional[str] = None # or empty str
    if (wikidump_dir / "siteinfo.json").exists():
        with open(wikidump_dir / "siteinfo.json", "r", encoding="utf-8") as f:
            siteinfo: Dict = json.load(f)

        general = siteinfo.get("query", {}).get("general", {})
        rightsinfo = siteinfo.get("query", {}).get("rightsinfo", {})


        sitename = general.get("sitename", None)
        assert isinstance(sitename, str) or sitename is None

        base_url = general.get("base", None)
        assert isinstance(base_url, str) or base_url is None
        if base_url:
            if base_url.startswith("//"):
                print(f"WARNING: base_url {base_url} starts with // (protocol-relative URLs), will convert to https://")
                # Convert protocol-relative URLs
                base_url = re.sub(r"^//", r"https://", base_url)

        logo_url = general.get("logo", None)

        lang = general.get("lang", None)
        assert isinstance(lang, str) or lang is None

        rights_text = rightsinfo.get("text", None)
        assert isinstance(rights_text, str) or rights_text is None

        rights_url = rightsinfo.get("url", None)
        if rights_url and "www.fandom.com" in rights_url and "/licensing" in rights_url:
            # Link the default license directly instead
            rights_url = "https://creativecommons.org/licenses/by-sa/3.0/"

        assert isinstance(rights_url, str) or rights_url is None

    if config.xml:
        xml_file_path = wikidump_dir / get_xml_filename(config)
        assert xml_file_path.exists()
        with open(xml_file_path, "rb") as f:
            xmlheader = f.read(1024 * 1024) # 1MiB
        # get sitename from xmlheader
        if not sitename and b"<sitename>" in xmlheader:
            sitename = xmlheader.split(b"<sitename>", 1)[1].split(b"</sitename>", 1)[0].decode("utf-8")
        if not base_url and b"<base>" in xmlheader:
            base_url = xmlheader.split(b"<base>", 1)[1].split(b"</base>", 1)[0].decode("utf-8")
    
    if not base_url:
        base_url = re.sub(r"(?im)/api\.php", r"", config.api or config.index)

    keywords = [
        "wiki",
        "wikiteam",
        "wikiteam3",
        "MediaWiki",
        wiki_prefix,
    ]
    if sitename:
        keywords.append(sitename)
    if not rights_url and not rights_text:
        keywords.append("unknowncopyright")

    licenseurl: Optional[str] = urllib.parse.urljoin(config.api or config.index, rights_url) if rights_url else None
    description =  f'<a href="{base_url}">{sitename or wiki_prefix}</a> dumped with <a href="https://github.com/saveweb/wikiteam3/" rel="nofollow">wikiteam3</a> tools.'

    metadata = {
        "mediatype": "web",
        "collection": arg.collection,
        "title": "Wiki - " + (sitename or wiki_prefix),
        "description": description, # without URL, to bypass IA's anti-spam.
        "language": lang,
        "last-updated-date": time.strftime("%Y-%m-%d", time.gmtime()),
        "subject": "; ".join(
            keywords
        ),  # Keywords should be separated by ; but it doesn't matter much; the alternative is to set one per field with subject[0], subject[1], ...
        "licenseurl": licenseurl or None,
        "rights": rights_text or None,
        "originalurl": config.api or config.index,
        "upload-state": "uploading",
        "scanner": f"wikiteam3 (v{getVersion()})",
    }
    print("=== Item metadata: ===")
    print(json.dumps(metadata, indent=4, sort_keys=True, ensure_ascii=False))
    print(f"logo_url: {logo_url}")

    return metadata, logo_url

def upload(arg: Args):
    ZstdCompressor()
    SevenZipCompressor()
    ia_keys = read_ia_keys(arg.keys_file)
    wikidump_dir = arg.wikidump_dir
    wikidump_dir.name # {prefix}-{wikidump_dumpdate}-wikidump (e.g. wiki.example.org-20230730-wikidump)
    assert wikidump_dir.name.endswith("-wikidump"), f"Expected wikidump_dir to end with -wikidump, got {wikidump_dir.name}"

    print(f"=== Loading config from {wikidump_dir} ===")

    # load config
    init_config = Config()
    init_config.path = str(wikidump_dir)
    config = load_config(config_filename="config.json", config=init_config)

    config.path = str(wikidump_dir) # override path

    print(config)

    assert wikidump_dir == Path(config.path).resolve()

    assert is_markfile_exists(config, ALL_DUMPED_MARK), "Imcomplete dump"
    if is_markfile_exists(config, UPLOADED_MARK):
        print(f"Already uploaded to IA ({UPLOADED_MARK} exists), bye!")
        return

    wikidump_dumpdate = wikidump_dir.name.split("-")[-2]
    assert config.date == wikidump_dumpdate
    if (not wikidump_dumpdate.isdigit()) or (not 20230730 < int(wikidump_dumpdate) < 9999_99_99):
        raise ValueError(f"Expected wikidump_dumpdate to be an 8-digit number, got {wikidump_dumpdate}")
    try:
        datetime.strptime(wikidump_dumpdate, "%Y%m%d")
    except ValueError as e:
        raise ValueError(f"Expected wikidump_dumpdate to be a valid date, got {wikidump_dumpdate}") from e

    # NOTE: Punycoded domain may contain multiple `-`
    # e.g. `xn--6qq79v.xn--rhqv96g-20230730-wikidump` (你好.世界_美丽-20230730-wikidump)
    _identifier = IDENTIFIER_PREFIX + url2prefix_from_config(config=config) + "-" + config.date
    identifier = IDENTIFIER_PREFIX + wikidump_dir.name.rstrip("-wikidump")
    assert identifier == _identifier

    item = get_item(identifier)

    print("=== Preparing files to upload ===")
    filedict = prepare_files_to_upload(wikidump_dir, config, item, parallel=arg.parallel, zstd_level=arg.zstd_level)

    print("=== Preparing metadata ===")
    metadata, logo_url = prepare_item_metadata(wikidump_dir, config, arg)

    print("=== Checking IA S3 load average (optional) ===")

    try:
        avg_load = ia_s3_tasks_load_avg(session=item.session) # check IA load
        print(f"IA S3 load: {avg_load * 100:.4f}%")
        if avg_load > 0.99:
            print("WARNING: IA S3 is heavily overloaded, upload may fail")
            print("Deciding whether to continue even if IA S3 is heavily overloaded... (20% chance to continue, random)")
            if random.random() < 0.8:
                print("To prevent IA S3 from being overloaded further, please try uploading later, exiting...")
                sys.exit(99)
            print("Continuing anyway...")
        elif avg_load > 0.9:
            print("WARNING: IA S3 is overloaded, upload may fail") 
    except Exception as e:
        traceback.print_exc()
        print(f"Failed to get IA S3 load average: {e}")
        print("Don't worry, it's optional.")


    if arg.dry_run:
        print("=== Dry run, exiting ===")
        return

    print("=== Uploading ===")
    upload_main_resouces(item, filedict, metadata, ia_keys)

    item = get_item(identifier)
    if logo_url:
        print("=== Uploading logo (optional) ===")
        try:
            logo_url = urllib.parse.urljoin(config.api or config.index, logo_url)
            upload_logo(item, logo_url, ia_keys)
        except Exception as e:
            traceback.print_exc()
            print(f"Failed to upload logo: {e}")
            print("Don't worry, it's optional.")
    
    item = get_item(identifier)
    print("=== Updating upload-state ===")
    if item.metadata.get("upload-state") != "uploaded":
        r = item.modify_metadata({"upload-state": "uploaded"}, access_key=ia_keys.access, secret_key=ia_keys.secret)
        assert isinstance(r, requests.Response)
        print(r.text)
        r.raise_for_status()
    print("=== Uploading complete ===")
    print(f"identifier: {identifier}")
    print(f"URL: https://archive.org/details/{identifier}")
    mark_as_done(config, UPLOADED_MARK, msg=f"identifier: {identifier}")

def upload_logo(item: Item, logo_url: str, ia_keys: IAKeys):
    assert logo_url
    assert item.identifier

    parsed_url = urllib.parse.urlparse(logo_url)
    logo_suff = parsed_url.path.split(".")[-1].lower()
    if len(logo_suff) >= 7:
        logo_suff = "unknown"
    logo_name = item.identifier + "_logo." + logo_suff
    for file_ in item.files:
        if file_["name"] == logo_name:
            print(f"Logo {logo_name} already exists, skip")
            return
    logo_io = None
    for tries_left in range(4, 0, -1):
        try:
            logo_io = BytesIO(requests.get(logo_url, timeout=20).content)
            break
        except Exception:
            if tries_left == 1:
                raise
            print(f"Failed to download logo, retrying ({tries_left} tries left)")
            time.sleep(3)

    assert logo_io

    r_co = item.upload(
        {logo_name: logo_io},
        access_key=ia_keys.access,
        secret_key=ia_keys.secret,
        verbose=True,
    )
    for r_resp in r_co:
        assert isinstance(r_resp, requests.Response)
        print(r_resp.text)
        r_resp.raise_for_status()

def upload_main_resouces(item: Item, filedict: Dict[str, str], metadata: Dict, ia_keys: IAKeys):
    r_co = item.upload(
        files=filedict,
        metadata=metadata,
        access_key=ia_keys.access,
        secret_key=ia_keys.secret,
        verbose=True,
        queue_derive=False, # disable derive
    )
    for r_resp in r_co:
        assert isinstance(r_resp, requests.Response)
        print(r_resp.text)
        r_resp.raise_for_status()
    print(f"Uploading {len(filedict)} files: Done.\n")

    identifier = item.identifier
    assert identifier

    item = get_item(identifier) # refresh item
    tries = 400
    for tries_left in range(tries, 0, -1):
        if item.exists:
            break

        print(f"Waiting for item to be created ({tries_left} tries left)  ...", end='\r')
        if tries < 395:
            print(f"IA overloaded, still waiting for item to be created ({tries_left} tries left)  ...", end='\r')
        time.sleep(30)
        item = get_item(identifier)

    if not item.exists:
        raise TimeoutError(f"IA overloaded, item still not created after {400 * 30} seconds")

def main():
    parser = argparse.ArgumentParser(
        """ Upload wikidump to the Internet Archive."""
    )

    parser.add_argument("-kf", "--keys_file", default="~/.wikiteam3_ia_keys.txt", dest="keys_file",
                        help="Path to the IA S3 keys file. (first line: access key, second line: secret key)"
                             " [default: ~/.wikiteam3_ia_keys.txt]")
    parser.add_argument("-c", "--collection", default=DEFAULT_COLLECTION, choices=[DEFAULT_COLLECTION, TEST_COLLECTION, WIKITEAM_COLLECTION])
    parser.add_argument("--dry-run", action="store_true", help="Dry run, do not upload anything.")
    parser.add_argument("-u", "--update", action="store_true",
                        help="Update existing item. [!! not implemented yet !!]")
    parser.add_argument("--bin-zstd", default="zstd", dest="bin_zstd",
                        help="Path to zstd binary. [default: zstd] "
                        "[!! not implemented yet !!]"
                        )
    parser.add_argument("--zstd-level", default=ZstdCompressor.DEFAULT_LEVEL, type=int, choices=range(17, 23),
                        help=f"Zstd compression level. [default: {ZstdCompressor.DEFAULT_LEVEL}] "
                        f"If you have a lot of RAM, recommend to use max level (22)."
                        )
    parser.add_argument("--bin-7z", default="7z", dest="bin_7z",
                        help="Path to 7z binary. [default: 7z] "
                        "[!! not implemented yet !!]"
                        )
    parser.add_argument("--parallel", action="store_true", help="Parallelize compression tasks")
    parser.add_argument("wikidump_dir")
    
    arg = Args(**vars(parser.parse_args()))
    print(arg)
    upload(arg)
    



if __name__ == "__main__":
    main()



import json
import os

import requests

from wikiteam3.dumpgenerator.cli import Delay
from wikiteam3.dumpgenerator.api import get_JSON
from wikiteam3.dumpgenerator.config import Config


def save_siteinfo(config: Config, session: requests.Session):
    """Save a file with site info"""

    assert config.api

    if os.path.exists("%s/siteinfo.json" % (config.path)):
        print("siteinfo.json exists, do not overwrite")
        return
    
    print("Downloading site info as siteinfo.json")

    # MediaWiki 1.13+
    r = session.get(
        url=config.api,
        params={
            "action": "query",
            "meta": "siteinfo",
            "siprop": "general|namespaces|statistics|dbrepllag|interwikimap|namespacealiases|specialpagealiases|usergroups|extensions|skins|magicwords|fileextensions|rightsinfo",
            "sinumberingroup": 1,
            "format": "json",
        },
        timeout=10,
    )
    # MediaWiki 1.11-1.12
    if not "query" in get_JSON(r):
        r = session.get(
            url=config.api,
            params={
                "action": "query",
                "meta": "siteinfo",
                "siprop": "general|namespaces|statistics|dbrepllag|interwikimap",
                "format": "json",
            },
            timeout=10,
        )
    # MediaWiki 1.8-1.10
    if not "query" in get_JSON(r):
        r = session.get(
            url=config.api,
            params={
                "action": "query",
                "meta": "siteinfo",
                "siprop": "general|namespaces",
                "format": "json",
            },
            timeout=10,
        )
    result = get_JSON(r)
    Delay(config=config)
    with open(
        "%s/siteinfo.json" % (config.path), "w", encoding="utf-8"
    ) as outfile:
        outfile.write(json.dumps(result, indent=4, sort_keys=True, ensure_ascii=False))



from typing import Dict, Optional
import xml.etree.ElementTree as ET

from lxml import etree
from lxml.builder import E

from wikiteam3.dumpgenerator.exceptions import PageMissingError

def make_xml_page_from_raw(xml: str, arvcontinue: Optional[str] = None) -> str:
    """Discard the metadata around a <page> element in <mediawiki> string

    arvcontinue: None -> disable arvcontinue (default)
    arvcontinue: string (including empty "") -> write arvcontinue to XML (for api:allrevisions resuming)
    """
    tree: ET.Element = ET.XML(xml)
    page: ET.Element | None = tree.find(".//{*}page")

    assert page is not None

    if arvcontinue is not None:
        page.attrib['arvcontinue'] = arvcontinue
    # remove namespace prefix
    for elem in tree.iter():
        elem.tag = elem.tag.split('}', 1)[-1]

    return ET.tostring(page, encoding="unicode", method="xml", xml_declaration=False)


def make_xml_from_page(page: Dict, arvcontinue: Optional[str] = None) -> str:
    """Output an XML document as a string from a page as in the API JSON

    arvcontinue: None -> disable arvcontinue (default)
    arvcontinue: string (including empty "") -> write arvcontinue to XML (for api:allrevisions resuming)
    """
    try:
        p = E.page(
            E.title(str(page["title"])),
            E.ns(str(page["ns"])),
            E.id(str(page["pageid"])),
        )
        if arvcontinue is not None:
            p.attrib['arvcontinue'] = arvcontinue
        for rev in page["revisions"]:
            # Older releases like MediaWiki 1.16 do not return all fields.
            if "userid" in rev:
                userid = rev["userid"]
            else:
                userid = 0
            if "size" in rev:
                size = rev["size"]
            else:
                size = 0

            # Create rev object
            revision = [E.id(str(rev["revid"])),
                E.timestamp(rev["timestamp"]),]

            # The text, user, comment, sha1 may be deleted/suppressed
            if (('texthidden' in rev) or ('textmissing' in rev)):
                print("Warning: text missing/hidden in pageid %d revid %d" % (page['pageid'], rev['revid']))
                revision.append(E.text(**{
                    'bytes': str(size),
                    'deleted': 'deleted',
                }))
            else:
                text = str(rev["*"])
                revision.append(E.text(text, **{
                    'bytes': str(size),
                    '{http://www.w3.org/XML/1998/namespace}space': 'preserve',
                }))

            if "user" not in rev:
                if "userhidden" not in rev:
                    print("Warning: user not hidden but missing user in pageid %d revid %d" % (page['pageid'], rev['revid']))
                revision.append(E.contributor(deleted="deleted"))
            else:
                revision.append(
                    E.contributor(
                        E.username(str(rev["user"])),
                        E.id(str(userid)),
                    )
                )

            if "sha1" not in rev:
                if "sha1hidden" in rev:
                    revision.append(E.sha1()) # stub
                else:
                    # The sha1 may not have been backfilled on older wikis or lack for other reasons (Wikia).
                    pass
            elif "sha1" in rev:
                revision.append(E.sha1(rev["sha1"]))


            if 'commenthidden' in rev:
                revision.append(E.comment(deleted="deleted"))
            elif "comment" in rev and rev["comment"]:
                revision.append(E.comment(str(rev["comment"])))

            if "contentmodel" in rev:
                revision.append(E.model(rev["contentmodel"]))
            if "contentformat" in rev:
                revision.append(E.format(rev["contentformat"]))
            if "parentid" in rev and int(rev["parentid"]) > 0:
                revision.append(E.parentid(str(rev["parentid"])))

            if "minor" in rev:
                revision.append(E.minor())

            # mwcli's dump.xml order
            revisionTags = ['id', 'parentid', 'timestamp', 'contributor', 'minor', 'comment', 'origin', 'model', 'format', 'text', 'sha1']
            revisionElementsDict = {elem.tag: elem for elem in revision}
            _revision = E.revision()
            for tag in revisionTags:
                if tag in revisionElementsDict:
                    _revision.append(revisionElementsDict.pop(tag))
            for elem in revisionElementsDict.values():
                _revision.append(elem)
            p.append(_revision)
    except KeyError as e:
        print(e)
        raise PageMissingError(page["title"], e)
    return etree.tostring(p, pretty_print=True, encoding="unicode")



import datetime

from wikiteam3.dumpgenerator.version import getVersion


def welcome():
    """Opening message"""

    welcome_string = f"# Welcome to DumpGenerator {getVersion()} by WikiTeam3 (GPL v3)"
    welcome_string += " " * (73 - len(welcome_string) - 1) + "#"
    copyright_string = f"# Copyright (C) 2011-{datetime.datetime.utcnow().year} WikiTeam developers"
    copyright_string += " " * (73 - len(copyright_string) - 1) + "#"

    return f"""\
#########################################################################
{welcome_string}
# More info at: <https://github.com/saveweb/wikiteam3>                  #
{copyright_string}
#########################################################################
"""


def bye(wikidump_dir = None):
    """Closing message"""
    print(
f"""
---> Done <---

If this is a public wiki, please, consider publishing this dump to the Internet Archive:

`wikiteam3uploader {wikidump_dir if wikidump_dir else ''}`

Good luck! Bye!
"""
)


import os

from wikiteam3.dumpgenerator.cli import Delay
from wikiteam3.utils import remove_IP
from wikiteam3.dumpgenerator.config import Config

def save_IndexPHP(config: Config, session):
    """Save index.php as .html, to preserve license details available at the botom of the page"""

    assert config.index

    if os.path.exists("%s/index.html" % (config.path)):
        print("index.html exists, do not overwrite")
    else:
        print("Downloading index.php (Main Page) as index.html")
        try:
            r = session.post(url=config.index, params=None, timeout=10)
            raw = str(r.text)
        except Exception as e:
            print("Error: %s" % (e))
            return
        Delay(config=config)
        raw = remove_IP(raw=raw)
        with open("%s/index.html" % (config.path), "w", encoding="utf-8") as outfile:
            outfile.write(raw)



import re
import time
import traceback
from typing import Dict, Optional
import xml.etree.ElementTree as ET
import xml.dom.minidom as MD

import requests

from wikiteam3.dumpgenerator.api import handle_StatusCode
from wikiteam3.dumpgenerator.config import Config
from wikiteam3.dumpgenerator.exceptions import PageMissingError, ExportAbortedError
from wikiteam3.dumpgenerator.log import log_error


def reconstructRevisions(root: ET.Element):
    #print ET.tostring(rev)
    page = ET.Element('stub')
    edits = 0

    if root.find('query').find('pages').find('page').find('revisions') is None: # type: ignore
        # case: https://wiki.archlinux.org/index.php?title=Arabic&action=history
        # No matching revisions were found.
        print('!!! No revisions found in page !!!')
        return page, edits # no revisions

    for rev in root.find('query').find('pages').find('page').find('revisions').findall('rev'): # type: ignore
        try:
            rev_ = ET.SubElement(page,'revision')
            # id
            ET.SubElement(rev_,'id').text = rev.attrib['revid']
            # parentid (optional, export-0.7+, positiveInteger)
            if 'parentid' in rev.attrib and int(rev.attrib['parentid']) > 0:
                ET.SubElement(rev_,'parentid').text = rev.attrib['parentid'] 
            # timestamp
            ET.SubElement(rev_,'timestamp').text = rev.attrib['timestamp']
            # contributor
            contributor = ET.SubElement(rev_,'contributor')
            if 'userhidden' not in rev.attrib:
                ET.SubElement(contributor,'username').text = rev.attrib['user']
                ET.SubElement(contributor,'id').text = rev.attrib['userid']
            else:
                contributor.set('deleted','deleted')
            # comment (optional)
            if 'commenthidden' in rev.attrib:
                print('commenthidden')
                comment = ET.SubElement(rev_,'comment')
                comment.set('deleted','deleted')
            elif 'comment' in rev.attrib and rev.attrib['comment']: # '' is empty
                comment = ET.SubElement(rev_,'comment')
                comment.text = rev.attrib['comment']
            else:
                # no comment or empty comment, do not create comment element
                pass

            # minor edit (optional)
            if 'minor' in rev.attrib:
                ET.SubElement(rev_,'minor')
            # model and format (optional, export-0.8+)
            if 'contentmodel' in rev.attrib:
                ET.SubElement(rev_,'model').text = rev.attrib['contentmodel'] # default: 'wikitext'
            if 'contentformat' in rev.attrib:
                ET.SubElement(rev_,'format').text = rev.attrib['contentformat'] # default: 'text/x-wiki'
            # text
            text = ET.SubElement(rev_,'text')
            if 'texthidden' not in rev.attrib:
                text.attrib['xml:space'] = "preserve"
                text.attrib['bytes'] = rev.attrib['size']
                text.text = rev.text
            else:
                # NOTE: this is not the same as the text being empty
                text.set('deleted','deleted')
            # sha1
            if 'sha1' not in rev.attrib:
                if 'sha1hidden' in rev.attrib:
                    ET.SubElement(rev_,'sha1') # stub
                else:
                    # The sha1 may not have been backfilled on older wikis or lack for other reasons (Wikia).
                    pass
            elif 'sha1' in rev.attrib:
                sha1 = ET.SubElement(rev_,'sha1')
                sha1.text = rev.attrib['sha1']
            
            edits += 1
        except Exception as e:
            #logerror(config=config, text='Error reconstructing revision, xml:%s' % (ET.tostring(rev)))
            print(ET.tostring(rev))
            traceback.print_exc()
            page = None
            edits = 0
            raise e
    return page,edits

def getXMLPageCoreWithApi(config: Config, session: requests.Session, params: Dict, headers: Optional[Dict]=None):
    """  """
    # just send the API request
    # if it fails, it will reduce params['rvlimit']
    xml = ''
    c = 0
    maxseconds = 100  # max seconds to wait in a single sleeping
    maxretries = config.retries  # x retries and skip
    increment = 20  # increment every retry

    while not re.search(r'</api>' if not config.curonly else r'</mediawiki>', xml) or re.search(r'</error>', xml):
        if c > 0 and c < maxretries:
            wait = increment * c < maxseconds and increment * \
                   c or maxseconds  # incremental until maxseconds
            print('    In attempt %d, XML for "%s" is wrong. Waiting %d seconds and reloading...' % (
            c, params['titles' if config.xmlapiexport else 'pages'], wait))
            time.sleep(wait)
            # reducing server load requesting smallest chunks (if curonly then
            # rvlimit = 1 from mother function)
            if params['rvlimit'] > 1:
                rvlimit = int(params['rvlimit'] / 2)  # half
                params['rvlimit'] = rvlimit if rvlimit > 1 else 1
            print('    We have retried %d times' % (c))
            print('    MediaWiki error for "%s", network error or whatever...' % (
            params['titles' if config.xmlapiexport else 'pages']))
            # If it's not already what we tried: our last chance, preserve only the last revision...
            # config.curonly means that the whole dump is configured to save only the last,
            # params['curonly'] should mean that we've already tried this
            # fallback, because it's set by the following if and passed to
            # getXMLPageCore
            # TODO: save only the last version when failed
            print('    Saving in the errors log, and skipping...')
            log_error(
                config=config,
                text='Error while retrieving the last revision of "%s". Skipping.' %
                     (params['titles' if config.xmlapiexport else 'pages']))
            raise ExportAbortedError(config.index)
            return ''  # empty xml

        # FIXME HANDLE HTTP Errors HERE
        try:
            r = session.get(url=config.api, params=params, headers=headers)
            handle_StatusCode(r)
            xml = r.text
            # print xml
        except requests.exceptions.ConnectionError as e:
            print('    Connection error: %s' % (str(e.args[0])))
            xml = ''
        except requests.exceptions.ReadTimeout as e:
            print("    Read timeout: %s" % (str(e.args[0])))
            xml = ""
        c += 1
    return xml


def getXMLPageWithApi(config: Config, title="", verbose=True, *, session: requests.Session):
    """ Get the full history (or current only) of a page using API:Query
        if params['curonly'] is set, then using export&exportwrap to export
    """

    title_ = title
    title_ = re.sub(' ', '_', title_)
    # do not convert & into %26, title_ = re.sub('&', '%26', title_)
    # action=query&rvlimit=50&format=xml&prop=revisions&titles=TITLE_HERE
    # &rvprop=timestamp%7Cuser%7Ccomment%7Ccontent%7Cids%7Cuserid%7Csha1%7Csize
    # print 'current:%s' % (title_)
    if not config.curonly:
        params = {'titles': title_, 'action': 'query', 'format': 'xml',
                  'prop': 'revisions',
                  'rvprop': # rvprop: <https://www.mediawiki.org/wiki/API:Revisions#Parameter_history>
                            '|'.join([
                                'timestamp', 'user', 'comment', 'content', # MW v????
                                'ids', 'flags', 'size', # MW v1.11
                                'userid', # MW v1.17
                                'sha1', # MW v1.19
                                'contentmodel' # MW v1.21
                             ]),
                  'rvcontinue': None,
                  'rvlimit': config.api_chunksize
                  }
    else:
        params = {'titles': title_, 'action': 'query', 'format': 'xml', 'export': 1, 'exportnowrap': 1}
    # print 'params:%s' % (params)
    if not config.curonly:
        firstpartok = False
        lastcontinue = None
        numberofedits = 0
        ret = ''
        continueKey: Optional[str] = None

        retries_left = config.retries
        while True:
            if retries_left <= 0:
                raise RuntimeError("Retries exceeded")
            # in case the last request is not right, saving last time's progress
            if not firstpartok:
                lastcontinue = params.get(continueKey, None) if continueKey is not None else None

            xml = getXMLPageCoreWithApi(params=params, config=config, session=session)
            if xml == "":
                # just return so that we can continue, and getXMLPageCoreWithApi will log the error
                return
            try:
                root = ET.fromstring(xml.encode('utf-8'))
            except Exception as e:
                retries_left -= 1
                traceback.print_exc()
                print("Retrying...")
                continue
            try:
                retpage = root.find('query').find('pages').find('page') # type: ignore
            except Exception:
                retries_left -= 1
                traceback.print_exc()
                print("Retrying...")
                continue

            assert retpage is not None, "Should have a page"

            if 'missing' in retpage.attrib or 'invalid' in retpage.attrib:
                print('Page not found')
                raise PageMissingError(params['titles'], xml)
            if not firstpartok:
                try:
                    # build the firstpart by ourselves to improve the memory usage
                    ret = '  <page>\n'
                    ret += '    <title>%s</title>\n' % (retpage.attrib['title'])
                    ret += '    <ns>%s</ns>\n' % (retpage.attrib['ns'])
                    ret += '    <id>%s</id>\n' % (retpage.attrib['pageid'])
                except Exception:
                    firstpartok = False
                    retries_left -= 1
                    traceback.print_exc()
                    print("Retrying...")
                    continue
                else:
                    firstpartok = True
                    yield ret

            # find the continue key
            continueVal = None
            if root.find('continue') is not None:
                # uses continue.rvcontinue
                # MW 1.26+
                continueKey = 'rvcontinue'
                continueVal = root.find('continue').attrib['rvcontinue'] # type: ignore
            elif root.find('query-continue') is not None:
                revContinue = root.find('query-continue').find('revisions') # type: ignore
                assert revContinue is not None, "Should only have revisions continue"
                if 'rvcontinue' in revContinue.attrib:
                    # MW 1.21 ~ 1.25
                    continueKey = 'rvcontinue'
                    continueVal = revContinue.attrib['rvcontinue']
                elif 'rvstartid' in revContinue.attrib:
                    # TODO: MW ????
                    continueKey = 'rvstartid'
                    continueVal = revContinue.attrib['rvstartid']
                else:
                    # blindly assume the first attribute is the continue key
                    # may never happen
                    assert len(revContinue.attrib) > 0, "Should have at least one attribute"
                    for continueKey in revContinue.attrib.keys():
                        continueVal = revContinue.attrib[continueKey]
                        break
            if continueVal is not None:
                params[continueKey] = continueVal

            # build the revision tags
            try:
                ret = ''
                edits = 0

                # transform the revision
                rev_, edits = reconstructRevisions(root=root)
                numberofedits += edits
                xmldom = MD.parseString(b'<stub1>' + ET.tostring(rev_) + b'</stub1>')
                # convert it into text in case it throws MemoryError
                # delete the first three line and last two line,which is for setting the indent
                ret += ''.join(xmldom.toprettyxml(indent='  ').splitlines(True)[3:-2])
                yield ret
                if config.curonly or continueVal is None:  # no continue
                    break
            except Exception:
                retries_left -= 1
                traceback.print_exc()
                print("Retrying...")
                params['rvcontinue'] = lastcontinue
                ret = ''
        yield '  </page>\n'
        if numberofedits == 0:
            raise PageMissingError(title=title_, xml=xml)
    else: # curonly
        xml = getXMLPageCoreWithApi(params=params, config=config, session=session)
        if xml == "":
            raise ExportAbortedError(config.index)
        if "</page>" not in xml:
            raise PageMissingError(title_, xml)

        yield xml.split("</page>")[0]

        # just for looking good :)
        r_timestamp = r'<timestamp>([^<]+)</timestamp>'

        numberofedits = 0
        numberofedits += len(re.findall(r_timestamp, xml))

        yield "</page>\n"

    if verbose:
        if (numberofedits == 1):
            print('    %s, 1 edit' % (title.strip()))
        else:
            print('    %s, %d edits' % (title.strip(), numberofedits))



import re
import sys
from typing import Dict

from wikiteam3.dumpgenerator.config import Config

def avoid_WikiMedia_projects(config: Config=None, other: Dict=None):
    """Skip Wikimedia projects and redirect to the dumps website"""

    # notice about wikipedia dumps
    url = ""
    if config.api:
        url = url + config.api
    if config.index:
        url = url + config.index
    if re.findall(
        r"(?i)(wikipedia|wikisource|wiktionary|wikibooks|wikiversity|wikimedia|wikispecies|wikiquote|wikinews|wikidata|wikivoyage)\.org",
        url,
    ):
        print("PLEASE, DO NOT USE THIS SCRIPT TO DOWNLOAD WIKIMEDIA PROJECTS!")
        print("Download the dumps from http://dumps.wikimedia.org")
        if not other["force"]:
            print("Thanks!")
            sys.exit(2)



""" Available since MediaWiki 1.27. login to a wiki using username and password (API) """

from typing import Optional

import requests


def fetch_login_token(session: requests.Session, api: str) -> Optional[str]:
    """ fetch login token by API .(MediaWiki 1.27+)"""

    response = session.get(
        url=api,
        params={
            'action': "query",
            'meta': "tokens",
            'type': "login",
            'format': "json"})
    data = response.json()
    try:
        token = data['query']['tokens']['logintoken']
        if type(token) is str:
            return token
    except KeyError:
        print('fetch login token: Oops! Something went wrong -- ', data)
        return None


def client_login(api: str ,session: requests.Session, username: str, password: str) -> Optional[requests.Session]:
    """ login to a wiki using username and password. (MediaWiki 1.27+)"""

    login_token = fetch_login_token(session=session, api=api)
    if not login_token:
        return None

    response = session.post(url=api, data={
        'action': "clientlogin",
        'username': username,
        'password': password,
        'loginreturnurl': 'http://127.0.0.1:5000/',
        'logintoken': login_token,
        'format': "json"
    })

    data = response.json()

    try:
        if data['clientlogin']['status'] == 'PASS':
            print('client login: Success! Welcome, ' + data['clientlogin']['username'] + '!')
    except KeyError:
        print('client login: Oops! Something went wrong -- ', data)
        return None


    return session


def bot_login(api:str ,session: requests.Session, username: str, password: str) -> Optional[requests.Session]:
    """ login to a wiki using BOT's name and password. (MediaWiki 1.27+) """

    login_token = fetch_login_token(session=session, api=api)
    if not login_token:
        return None

    response = session.post(url=api, data={
        'action': "login",
        'lgname': username,
        'lgpassword': password,
        'lgtoken': login_token,
        'format': "json"
    })

    data = response.json()

    try:
        if data['login']['result'] == 'Success':
            print('bot login: Success! Welcome, ' + data['login']['lgusername'] + '!')
    except KeyError:
        print('bot login: Oops! Something went wrong -- ' + data)
        return None
    
    return session


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wikiteam3',
 'wikiteam3.dumpgenerator',
 'wikiteam3.dumpgenerator.api',
 'wikiteam3.dumpgenerator.cli',
 'wikiteam3.dumpgenerator.dump',
 'wikiteam3.dumpgenerator.dump.image',
 'wikiteam3.dumpgenerator.dump.misc',
 'wikiteam3.dumpgenerator.dump.page',
 'wikiteam3.dumpgenerator.dump.page.xmlexport',
 'wikiteam3.dumpgenerator.dump.page.xmlrev',
 'wikiteam3.dumpgenerator.dump.xmldump',
 'wikiteam3.dumpgenerator.log',
 'wikiteam3.uploader',
 'wikiteam3.utils',
 'wikiteam3.utils.login']

package_data = \
{'': ['*']}

install_requires = \
['file_read_backwards>=3.0.0,<4.0.0',
 'httpx>=0.24.1,<0.25.0',
 'internetarchive>=3.5.0,<4.0.0',
 'lxml>=4.9.2,<5.0.0',
 'mwclient>=0.10.1,<0.11.0',
 'python-slugify>=8.0.1,<9.0.0',
 'requests>=2.31.0,<3.0.0']

entry_points = \
{'console_scripts': ['dumpgenerator = wikiteam3.dumpgenerator:main_deprecated',
                     'wikiteam3dumpgenerator = wikiteam3.dumpgenerator:main',
                     'wikiteam3uploader = wikiteam3.uploader:main']}

setup_kwargs = {
    'name': 'wikiteam3',
    'version': '4.1.6',
    'description': 'Tools for downloading and preserving MediaWikis. We archive MediaWikis, from Wikipedia to tiniest wikis.',
    'long_description': '# `wikiteam3`\n\n<!-- !["MediaWikiArchive.png"](./MediaWikiArchive.png) -->\n<div align=center><img width = \'150\' height =\'150\' src ="./MediaWikiArchive.png"/></div>\n\n> Countless MediaWikis are still waiting to be archived.\n>\n> _Image by [@gledos](https://github.com/gledos/)_\n\n`wikiteam3` is a fork of `mediawiki-scraper`.\n\n<details>\n\n## Why we fork mediawiki-scraper\n\nOriginally, mediawiki-scraper was named wikiteam3, but wikiteam upstream (py2 version) suggested that the name should be changed to avoid confusion with the original wikiteam.  \nHalf a year later, we didn\'t see any py3 porting progress in the original wikiteam, and mediawiki-scraper lacks "code" reviewers.  \nSo, we decided to break that suggestion, fork and named it back to wikiteam3, put the code here, and release it to pypi wildly.\n\nEverything still under GPLv3 license.\n\n</details>\n\n## Installation/Upgrade\n\n```shell\npip install wikiteam3 --upgrade\n```\n\n## Dumpgenerator usage\n\n### Downloading a wiki with complete XML history and images\n\n```bash\nwikiteam3dumpgenerator http://wiki.domain.org --xml --images\n```\n\n>[!WARNING]\n>\n> `NTFS/Windows` users please note: When using `--images`, because NTFS does not allow characters such as `:*?"<>|` in filenames, some files may not be downloaded, please pay attention to the `XXXXX could not be created by OS` error in your `errors.log`.\n> We will not make special treatment for NTFS/EncFS "path too long/illegal filename", highly recommend you to use ext4/xfs/btrfs, etc.\n> <details>\n> - Introducing the "illegal filename rename" mechanism will bring complexity. WikiTeam(python2) had this before, but it caused more problems, so it was removed in WikiTeam3.\n> - It will cause confusion to the final user of wikidump (usually the Wiki site administrator).\n> - NTFS is not suitable for large-scale image dump with millions of files in a single directory.(Windows background service will occasionally scan the whole disk, we think there should be no users using WIN/NTFS to do large-scale MediaWiki archive)\n> - Using other file systems can solve all problems.\n> </details>\n\n### Manually specifying `api.php` and/or `index.php`\n\nIf the script can\'t find itself the `api.php` and/or `index.php` paths, then you can provide them:\n\n```bash\nwikiteam3dumpgenerator --api http://wiki.domain.org/w/api.php --xml --images\n```\n\n```bash\nwikiteam3dumpgenerator --api http://wiki.domain.org/w/api.php --index http://wiki.domain.org/w/index.php \\\n    --xml --images\n```\n\nIf you only want the XML histories, just use `--xml`. For only the images, just `--images`. For only the current version of every page, `--xml --curonly`.\n\n### Resuming an incomplete dump\n\n<details>\n\n```bash\nwikiteam3dumpgenerator \\\n    --api http://wiki.domain.org/w/api.php --xml --images --resume --path /path/to/incomplete-dump\n```\n\nIn the above example, `--path` is only necessary if the download path (wikidump dir) is not the default.\n\n>[!NOTE]\n>\n> en: When resuming an incomplete dump, the configuration in `config.json` will override the CLI parameters. (But not all CLI parameters will be ignored, check `config.json` for details)\n\n`wikiteam3dumpgenerator` will also ask you if you want to resume if it finds an incomplete dump in the path where it is downloading.\n\n</details>\n\n## Using `wikiteam3uploader`\n\n### Requirements\n\n> [!NOTE]\n>\n> Please make sure you have the following requirements before using `wikiteam3uploader`, and you don\'t need to install them if you don\'t wanna upload the dump to IA.\n\n- unbinded localhost port 62954 (for multiple processes compressing queue)\n- 3GB+ RAM (~2.56GB for commpressing)\n- 64-bit OS (required by 2G `wlog` size)\n\n- `7z` (binary)\n    > Debian/Ubuntu: install `p7zip-full`  \n\n    > [!NOTE]\n    >\n    > Windows: install <https://7-zip.org> and add `7z.exe` to PATH\n- `zstd` (binary)\n    > 1.5.5+ (recommended), v1.5.0-v1.5.4(DO NOT USE), 1.4.8 (minimum)  \n    > install from <hhttps://github.com/facebook/zstd>  \n\n    > [!NOTE]\n    >\n    > Windows: add `zstd.exe` to PATH\n\n### Uploader usage\n\n> [!NOTE]\n>\n> Read `wikiteam3uploader --help` and do not forget `~/.wikiteam3_ia_keys.txt` before using `wikiteam3uploader`.\n\n```bash\nwikiteam3uploader {YOUR_WIKI_DUMP_PATH}\n```\n\n## Checking dump integrity\n\nTODO: xml2titles.py\n\nIf you want to check the XML dump integrity, type this into your command line to count title, page and revision XML tags:\n\n```bash\ngrep -E \'<title(.*?)>\' *.xml -c; grep -E \'<page(.*?)>\' *.xml -c; grep \\\n    "</page>" *.xml -c;grep -E \'<revision(.*?)>\' *.xml -c;grep "</revision>" *.xml -c\n```\n\nYou should see something similar to this (not the actual numbers) - the first three numbers should be the same and the last two should be the same as each other:\n\n```bash\n580\n580\n580\n5677\n5677\n```\n\nIf your first three numbers or your last two numbers are different, then, your XML dump is corrupt (it contains one or more unfinished ```</page>``` or ```</revision>```). This is not common in small wikis, but large or very large wikis may fail at this due to truncated XML pages while exporting and merging. The solution is to remove the XML dump and re-download, a bit boring, and it can fail again.\n\n## import wikidump to MediaWiki / wikidump data tips\n\n> [!IMPORTANT]\n>\n> In the article name, spaces and underscores are treated as equivalent and each is converted to the other in the appropriate context (underscore in URL and database keys, spaces in plain text). <https://www.mediawiki.org/wiki/Manual:Title.php#Article_name>\n\n> [!NOTE]\n>\n> `WikiTeam3` uses `zstd` to compress `.xml` and `.txt` files, and `7z` to pack images (media files).  \n> `zstd` is a very fast stream compression algorithm, you can use `zstd -d` to decompress `.zst` file/steam.\n\n## Contributors\n\n**WikiTeam** is the [Archive Team](http://www.archiveteam.org) [[GitHub](https://github.com/ArchiveTeam)] subcommittee on wikis.\nIt was founded and originally developed by [Emilio J. Rodríguez-Posada](https://github.com/emijrp), a Wikipedia veteran editor and amateur archivist. Thanks to people who have helped, especially to: [Federico Leva](https://github.com/nemobis), [Alex Buie](https://github.com/ab2525), [Scott Boyd](http://www.sdboyd56.com), [Hydriz](https://github.com/Hydriz), Platonides, Ian McEwen, [Mike Dupont](https://github.com/h4ck3rm1k3), [balr0g](https://github.com/balr0g) and [PiRSquared17](https://github.com/PiRSquared17).\n\n**Mediawiki-Scraper** The Python 3 initiative is currently being led by [Elsie Hupp](https://github.com/elsiehupp), with contributions from [Victor Gambier](https://github.com/vgambier), [Thomas Karcher](https://github.com/t-karcher), [Janet Cobb](https://github.com/randomnetcat), [yzqzss](https://github.com/yzqzss), [NyaMisty](https://github.com/NyaMisty) and [Rob Kam](https://github.com/robkam)\n\n**WikiTeam3** Every archivist who has uploaded a wikidump to the [Internet Archive](https://archive.org/search?query=subject%3Awikiteam3).\n',
    'author': 'yzqzss',
    'author_email': 'yzqzss@yandex.com',
    'maintainer': 'yzqzss',
    'maintainer_email': 'yzqzss@yandex.com',
    'url': 'https://github.com/saveweb/wikiteam3',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
